# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2018 Mellanox Technologies, Inc. All rights reserved. See COPYING file
# Copyright 2020 Amazon.com, Inc. or its affiliates. All rights reserved.
"""
Test module for pyverbs' device module.
"""
from multiprocessing import Process, Queue
import unittest
import resource
import random
import errno
import os

from pyverbs.pyverbs_error import PyverbsError, PyverbsRDMAError
from tests.base import PyverbsAPITestCase
from pyverbs.device import Context, DM
import tests.utils as u
import pyverbs.device as d
import pyverbs.enums as e

PAGE_SIZE = resource.getpagesize()


class DeviceTest(PyverbsAPITestCase):
    """
    Test various functionalities of the Device class.
    """

    def get_device_list(self):
        lst = d.get_device_list()
        if len(lst) == 0:
            raise unittest.SkipTest('No IB device found')
        dev_name = self.config['dev']
        if dev_name:
            for dev in lst:
                if dev.name.decode() == dev_name:
                    lst = [dev]
                    break
            if len(lst) == 0:
                raise PyverbsRDMAError(f'No IB device with name {dev_name} found')
        return lst

    def test_dev_list(self):
        """
        Verify that it's possible to get IB devices list.
        """
        self.get_device_list()

    def test_open_dev(self):
        """
        Test ibv_open_device()
        """
        for dev in self.get_device_list():
            d.Context(name=dev.name.decode())

    def test_query_device(self):
        """
        Test ibv_query_device()
        """
        for dev in self.get_device_list():
            with d.Context(name=dev.name.decode()) as ctx:
                attr = ctx.query_device()
                self.verify_device_attr(attr, dev)

    def test_query_pkey(self):
        """
        Test ibv_query_pkey()
        """
        for dev in self.get_device_list():
            with d.Context(name=dev.name.decode()) as ctx:
                if dev.node_type == e.IBV_NODE_CA:
                    ctx.query_pkey(port_num=self.ib_port, index=0)

    def test_query_gid(self):
        """
        Test ibv_query_gid()
        """
        for dev in self.get_device_list():
            with d.Context(name=dev.name.decode()) as ctx:
                gid_tbl_len = ctx.query_port(self.ib_port).gid_tbl_len
                if gid_tbl_len > 0:
                    ctx.query_gid(port_num=self.ib_port, index=0)

    def test_query_gid_table(self):
        """
        Test ibv_query_gid_table()
        """
        devs = self.get_device_list()
        with d.Context(name=devs[0].name.decode()) as ctx:
            device_attr = ctx.query_device()
            max_entries = 0
            for port_num in range(1, device_attr.phys_port_cnt + 1):
                port_attr = ctx.query_port(port_num)
                max_entries += port_attr.gid_tbl_len
            try:
                if max_entries > 0:
                    ctx.query_gid_table(max_entries)
            except PyverbsRDMAError as ex:
                if ex.error_code in [-errno.EOPNOTSUPP, -errno.EPROTONOSUPPORT]:
                    raise unittest.SkipTest('ibv_query_gid_table is not'\
                                            ' supported on this device')
                raise ex

    def test_query_gid_table_bad_flow(self):
        """
        Test ibv_query_gid_table() with too small a buffer
        """
        try:
            self.ctx.query_gid_table(0)
        except PyverbsRDMAError as ex:
            if ex.error_code in [-errno.EOPNOTSUPP, -errno.EPROTONOSUPPORT]:
                raise unittest.SkipTest('ibv_query_gid_table is not'
                                        ' supported on this device')
            self.assertEqual(ex.error_code, -errno.EINVAL,
                             f'Got -{os.strerror(-ex.error_code)} but '
                             f'Expected -{os.strerror(errno.EINVAL)} ')
        else:
            raise PyverbsRDMAError('Successfully queried '
                                   'gid_table with an insufficient buffer')

    def test_query_gid_ex(self):
        """
        Test ibv_query_gid_ex()
        """
        devs = self.get_device_list()
        with d.Context(name=devs[0].name.decode()) as ctx:
            try:
                gid_tbl_len = ctx.query_port(self.ib_port).gid_tbl_len
                if gid_tbl_len > 0:
                    ctx.query_gid_ex(port_num=self.ib_port, gid_index=0)
            except PyverbsRDMAError as ex:
                if ex.error_code in [errno.EOPNOTSUPP, errno.EPROTONOSUPPORT]:
                    raise unittest.SkipTest('ibv_query_gid_ex is not'\
                                            ' supported on this device')
                raise ex

    def test_query_gid_ex_bad_flow(self):
        """
        Test ibv_query_gid_ex() with an empty index
        """
        try:
            port_attr = self.ctx.query_port(self.ib_port)
            max_entries = 0
            for port_num in range(1, self.attr.phys_port_cnt + 1):
                attr = self.ctx.query_port(port_num)
                max_entries += attr.gid_tbl_len
            if max_entries > 0:
                gid_indices = {gid_entry.gid_index for gid_entry in
                               self.ctx.query_gid_table(max_entries) if gid_entry.port_num == self.ib_port}
            else:
                gid_indices = {}

            possible_indices = set(range(port_attr.gid_tbl_len)) if port_attr.gid_tbl_len > 1 else set()
            try:
                no_gid_index = possible_indices.difference(gid_indices).pop()
            except KeyError:
                # all indices are populated by GIDs
                raise unittest.SkipTest('All gid indices populated,'
                                        ' cannot check bad flow')

            self.ctx.query_gid_ex(port_num=self.ib_port, gid_index=no_gid_index)
        except PyverbsRDMAError as ex:
            if ex.error_code in [errno.EOPNOTSUPP, errno.EPROTONOSUPPORT]:
                raise unittest.SkipTest('ibv_query_gid_ex is not'
                                        ' supported on this device')
            self.assertEqual(ex.error_code, errno.ENODATA,
                             f'Got {os.strerror(ex.error_code)} but '
                             f'Expected {os.strerror(errno.ENODATA)}')
        else:
            raise PyverbsRDMAError('Successfully queried '
                                   f'non-existent gid index {no_gid_index}')

    @staticmethod
    def verify_device_attr(attr, device):
        """
        Helper method that verifies correctness of some members of DeviceAttr
        object.
        :param attr: A DeviceAttr object
        :param device: A Device object
        :return: None
        """
        if device.node_type != e.IBV_NODE_UNSPECIFIED and device.node_type != e.IBV_NODE_UNKNOWN:
            assert attr.node_guid != 0
            assert attr.sys_image_guid != 0
        assert attr.max_mr_size > PAGE_SIZE
        assert attr.page_size_cap >= PAGE_SIZE
        assert attr.vendor_id != 0
        assert attr.max_qp > 0
        assert attr.max_qp_wr > 0
        assert attr.max_sge > 0
        assert attr.max_sge_rd >= 0
        assert attr.max_cq > 0
        assert attr.max_cqe > 0
        assert attr.max_mr > 0
        assert attr.max_pd > 0
        if device.node_type == e.IBV_NODE_CA:
            assert attr.max_pkeys > 0

    def test_query_device_ex(self):
        """
        Test ibv_query_device_ex()
        """
        for dev in self.get_device_list():
            with d.Context(name=dev.name.decode()) as ctx:
                attr_ex = ctx.query_device_ex()
                self.verify_device_attr(attr_ex.orig_attr, dev)

    def test_phys_port_cnt_ex(self):
        """
        Test phys_port_cnt_ex
        """
        for dev in self.get_device_list():
            with d.Context(name=dev.name.decode()) as ctx:
                attr_ex = ctx.query_device_ex()
                phys_port_cnt = attr_ex.orig_attr.phys_port_cnt
                phys_port_cnt_ex = attr_ex.phys_port_cnt_ex
                if phys_port_cnt_ex > 255:
                    self.assertEqual(phys_port_cnt, 255,
                                     f'phys_port_cnt should be 255 if ' +
                                     f'phys_port_cnt_ex is bigger than 255')
                else:
                    self.assertEqual(phys_port_cnt, phys_port_cnt_ex,
                                     f'phys_port_cnt_ex and phys_port_cnt ' +
                                     f'should be equal if number of ports is ' +
                                     f'less than 256')

    @staticmethod
    def verify_port_attr(attr):
        """
        Helper method that verifies correctness of some members of PortAttr
        object.
        :param attr: A PortAttr object
        :return: None
        """
        assert 'Invalid' not in d.phys_state_to_str(attr.state)
        assert 'Invalid' not in d.translate_mtu(attr.max_mtu)
        assert 'Invalid' not in d.translate_mtu(attr.active_mtu)
        assert 'Invalid' not in d.width_to_str(attr.active_width)
        assert 'Invalid' not in d.speed_to_str(attr.active_speed,
                                               attr.active_speed_ex)
        assert 'Invalid' not in d.translate_link_layer(attr.link_layer)
        assert attr.max_msg_sz > 0x1000

    def test_query_port(self):
        """
        Test ibv_query_port
        """
        for dev in self.get_device_list():
            with d.Context(name=dev.name.decode()) as ctx:
                port_attr = ctx.query_port(self.ib_port)
                self.verify_port_attr(port_attr)

    def test_query_port_bad_flow(self):
        """
        Verify that querying non-existing ports fails as expected
        """
        for dev in self.get_device_list():
            with d.Context(name=dev.name.decode()) as ctx:
                num_ports = ctx.query_device().phys_port_cnt
                try:
                    port = num_ports + random.randint(1, 10)
                    ctx.query_port(port)
                except PyverbsRDMAError as e:
                    assert 'Failed to query port' in e.args[0]
                    assert 'Invalid argument' in e.args[0]
                else:
                    raise PyverbsRDMAError(
                        'Successfully queried non-existing port {p}'. \
                        format(p=port))


class DMTest(PyverbsAPITestCase):
    """
    Test various functionalities of the DM class.
    """
    def setUp(self):
        super().setUp()
        if self.attr_ex.max_dm_size == 0:
            raise unittest.SkipTest('Device memory is not supported')

    def test_create_dm(self):
        """
        test ibv_alloc_dm()
        """
        dm_len = random.randrange(u.MIN_DM_SIZE, int(self.attr_ex.max_dm_size/2),
                                  u.DM_ALIGNMENT)
        dm_attrs = u.get_dm_attrs(dm_len)
        with d.DM(self.ctx, dm_attrs):
            pass

    def test_destroy_dm(self):
        """
        test ibv_free_dm()
        """
        dm_len = random.randrange(u.MIN_DM_SIZE, int(self.attr_ex.max_dm_size/2),
                                  u.DM_ALIGNMENT)
        dm_attrs = u.get_dm_attrs(dm_len)
        dm = d.DM(self.ctx, dm_attrs)
        dm.close()

    def test_create_dm_bad_flow(self):
        """
        test ibv_alloc_dm() with an illegal size and comp mask
        """
        dm_len = self.attr_ex.max_dm_size + 1
        dm_attrs = u.get_dm_attrs(dm_len)
        try:
            d.DM(self.ctx, dm_attrs)
        except PyverbsRDMAError as e:
            assert 'Failed to allocate device memory of size' in \
                   e.args[0]
            assert 'Max available size' in e.args[0]
        else:
            raise PyverbsError(
                'Created a DM with size larger than max reported')
        dm_attrs.comp_mask = random.randint(1, 100)
        try:
            d.DM(self.ctx, dm_attrs)
        except PyverbsRDMAError as e:
            assert 'Failed to allocate device memory of size' in \
                   e.args[0]
        else:
            raise PyverbsError(
                'Created a DM with illegal comp mask {c}'. \
                format(c=dm_attrs.comp_mask))

    def test_destroy_dm_bad_flow(self):
        """
        Test calling ibv_free_dm() twice
        """
        dm_len = random.randrange(u.MIN_DM_SIZE, int(self.attr_ex.max_dm_size/2),
                                  u.DM_ALIGNMENT)
        dm_attrs = u.get_dm_attrs(dm_len)
        dm = d.DM(self.ctx, dm_attrs)
        dm.close()
        dm.close()

    def test_dm_write(self):
        """
        Test writing to the device memory
        """
        dm_len = random.randrange(u.MIN_DM_SIZE, int(self.attr_ex.max_dm_size/2),
                                  u.DM_ALIGNMENT)
        dm_attrs = u.get_dm_attrs(dm_len)
        with d.DM(self.ctx, dm_attrs) as dm:
            data_length = random.randrange(4, dm_len, u.DM_ALIGNMENT)
            data_offset = random.randrange(0, dm_len - data_length,
                                           u.DM_ALIGNMENT)
            data = 'a' * data_length
            dm.copy_to_dm(data_offset, data.encode(), data_length)

    def test_dm_write_bad_flow(self):
        """
        Test writing to the device memory with bad offset and length
        """
        dm_len = random.randrange(u.MIN_DM_SIZE, int(self.attr_ex.max_dm_size/2),
                                  u.DM_ALIGNMENT)
        dm_attrs = u.get_dm_attrs(dm_len)
        with d.DM(self.ctx, dm_attrs) as dm:
            data_length = random.randrange(4, dm_len, u.DM_ALIGNMENT)
            data_offset = random.randrange(0, dm_len - data_length,
                                           u.DM_ALIGNMENT)
            data_offset += 1  # offset needs to be a multiple of 4
            data = 'a' * data_length
            try:
                dm.copy_to_dm(data_offset, data.encode(), data_length)
            except PyverbsRDMAError as e:
                assert 'Failed to copy to dm' in e.args[0]
            else:
                raise PyverbsError(
                    'Wrote to device memory with a bad offset')

    def test_dm_read(self):
        """
        Test reading from the device memory
        """
        dm_len = random.randrange(u.MIN_DM_SIZE, int(self.attr_ex.max_dm_size/2),
                                  u.DM_ALIGNMENT)
        dm_attrs = u.get_dm_attrs(dm_len)
        with d.DM(self.ctx, dm_attrs) as dm:
            data_length = random.randrange(4, dm_len, u.DM_ALIGNMENT)
            data_offset = random.randrange(0, dm_len - data_length,
                                           u.DM_ALIGNMENT)
            data = 'a' * data_length
            dm.copy_to_dm(data_offset, data.encode(), data_length)
            read_str = dm.copy_from_dm(data_offset, data_length)
            assert read_str.decode() == data

    def alloc_dm(self, res_queue, size):
        """
        Alloc device memory. Used by multiple processes that allocate DMs in
        parallel.
        :param res_queue: Result Queue to return the result to the parent
                            process.
        :param size: The DM allocation size.
        :return: None
        """
        try:
            d.DM(self.ctx, d.AllocDmAttr(length=size))
        except PyverbsError as err:
            res_queue.put(err.error_code)
        res_queue.put(0)

    def test_multi_process_alloc_dm(self):
        """
        Several processes try to allocate device memory simultaneously.
        """
        res_queue = Queue()
        processes = []
        processes_num = 5
        # Dividing the max dm size by 2 since we're not
        # guaranteed to have the max size free for us.
        total_size = self.attr_ex.max_dm_size / 2 / processes_num
        for i in range(processes_num):
            processes.append(Process(target=self.alloc_dm,
                                     args=(res_queue, total_size)))
        for i in range(processes_num):
            processes[i].start()
        for i in range(processes_num):
            processes[i].join()
            rc = res_queue.get()
            self.assertEqual(rc, 0, f'Parallel device memory allocation failed with errno: {rc}')


class SharedDMTest(PyverbsAPITestCase):
    """
    Tests shared device memory by importing DMs
    """
    def setUp(self):
        super().setUp()
        if self.attr_ex.max_dm_size == 0:
            raise unittest.SkipTest('Device memory is not supported')
        self.dm_size = int(self.attr_ex.max_dm_size / 2)

    def test_import_dm(self):
        """
        Creates a DM and imports it from a different (duplicated) Context.
        Then writes some data to the original DM, reads it from the imported DM
        and verifies that the read data is as expected.
        """
        with d.DM(self.ctx, d.AllocDmAttr(length=self.dm_size)) as dm:
            cmd_fd_dup = os.dup(self.ctx.cmd_fd)
            try:
                imported_ctx = Context(cmd_fd=cmd_fd_dup)
                imported_dm = DM(imported_ctx, handle=dm.handle)
            except PyverbsRDMAError as ex:
                if ex.error_code in [errno.EOPNOTSUPP, errno.EPROTONOSUPPORT]:
                    raise unittest.SkipTest('Some object imports are not supported')
                raise ex
            original_data = b'\xab' * self.dm_size
            dm.copy_to_dm(0, original_data, self.dm_size)
            read_data = imported_dm.copy_from_dm(0, self.dm_size)
            self.assertEqual(original_data, read_data)
            imported_dm.unimport()



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2021 Nvidia Inc. All rights reserved. See COPYING file

from libc.stdlib cimport calloc, free
from libc.stdint cimport uint8_t
from libc.string cimport memcpy
import weakref

from .pyverbs_error import PyverbsRDMAError, PyverbsError, PyverbsUserError
from pyverbs.base import PyverbsRDMAErrno
from pyverbs.base cimport close_weakrefs
cimport pyverbs.libibverbs_enums as e
from pyverbs.device cimport Context
from pyverbs.wr cimport RecvWR
from pyverbs.cq cimport CQ
from pyverbs.pd cimport PD
from pyverbs.qp cimport QP


cdef class WQInitAttr(PyverbsObject):
    def __init__(self, wq_context=None, PD wq_pd=None, CQ wq_cq=None, wq_type=e.IBV_WQT_RQ,
                 max_wr=100, max_sge=1, comp_mask=0, create_flags=0):
        """
        Initializes a WqInitAttr object representing ibv_wq_init_attr struct.
        :param wq_context: Associated WQ context
        :param wq_pd: PD to be associated with the WQ
        :param wq_cq: CQ to be associated with the WQ
        :param wp_type: The desired WQ type
        :param max_wr: Requested max number of outstanding WRs in the WQ
        :param max_sge: Requested max number of scatter/gather (s/g) elements per WR in the WQ
        :param comp_mask: Identifies valid fields
        :param create_flags: Creation flags for the WQ
        :return: A WqInitAttr object
        """
        super().__init__()
        self.attr.wq_context = <void*>(wq_context) if wq_context else NULL
        self.attr.wq_type = wq_type
        self.attr.max_wr = max_wr
        self.attr.max_sge = max_sge
        self.pd = wq_pd
        self.attr.pd = wq_pd.pd if wq_pd else NULL
        self.cq = wq_cq
        self.attr.cq = wq_cq.cq if wq_cq else NULL
        self.attr.comp_mask = comp_mask
        self.attr.create_flags = create_flags

    @property
    def wq_type(self):
        return self.attr.wq_type
    @wq_type.setter
    def wq_type(self, val):
        self.attr.wq_type = val

    @property
    def pd(self):
        return self.pd
    @pd.setter
    def pd(self, PD val):
        self.pd = val
        self.attr.pd = <v.ibv_pd*>val.pd

    @property
    def cq(self):
        return self.cq
    @cq.setter
    def cq(self, CQ val):
        self.cq = val
        self.attr.cq = <v.ibv_cq*>val.cq


cdef class WQAttr(PyverbsObject):
    def __init__(self, attr_mask=0, wq_state=0, curr_wq_state=0, flags=0, flags_mask=0):
        """
        Initializes a WQAttr object which represents ibv_wq_attr struct. It
        can be used to modify a WQ.
        :param attr_mask: Identifies valid fields
        :param wq_state: Desired WQ state
        :param curr_wq_state: Current WQ state
        :param flags: Flags values to modify
        :param flags_mask: Which flags to modify
        :return: An initialized WQAttr object
        """
        super().__init__()
        self.attr.attr_mask = attr_mask
        self.attr.wq_state = wq_state
        self.attr.curr_wq_state = curr_wq_state
        self.attr.flags = flags
        self.attr.flags_mask = flags_mask

    @property
    def wq_state(self):
        return self.attr.wq_state
    @wq_state.setter
    def wq_state(self, val):
        self.attr.wq_state = val

    @property
    def attr_mask(self):
        return self.attr.attr_mask
    @attr_mask.setter
    def attr_mask(self, val):
        self.attr.attr_mask = val

    @property
    def curr_wq_state(self):
        return self.attr.curr_wq_state
    @curr_wq_state.setter
    def curr_wq_state(self, val):
        self.attr.curr_wq_state = val

    @property
    def flags(self):
        return self.attr.flags
    @flags.setter
    def flags(self, val):
        self.attr.flags = val

    @property
    def flags_mask(self):
        return self.attr.flags_mask
    @flags_mask.setter
    def flags_mask(self, val):
        self.attr.flags_mask = val


cdef class WQ(PyverbsCM):
    def __init__(self, Context ctx, WQInitAttr attr):
        """
        Creates a WQ object.
        :param ctx: The context the wq will be associated with.
        :param attr: WQ initial attributes of type WQInitAttr.
        :return: A WQ object
        """
        super().__init__()
        self.wq = v.ibv_create_wq(ctx.context, &attr.attr)
        if self.wq == NULL:
            raise PyverbsRDMAErrno('Failed to create WQ')
        self.context = ctx
        ctx.add_ref(self)
        pd = <PD>attr.pd
        pd.add_ref(self)
        self.pd = pd
        cq = <CQ>attr.cq
        cq.add_ref(self)
        self.cq = cq
        self.rwq_ind_tables = weakref.WeakSet()

    cpdef add_ref(self, obj):
        if isinstance(obj, RwqIndTable):
            self.rwq_ind_tables.add(obj)
        else:
            raise PyverbsError('Unrecognized object type')

    def modify(self, WQAttr wq_attr not None):
        """
        Modify the WQ
        :param qp_attr: A WQAttr object with updated values to be applied to
                        the WQ
        :return: None
        """
        rc = v.ibv_modify_wq(self.wq, &wq_attr.attr)
        if rc != 0:
            raise PyverbsRDMAError('Failed to modify WQ', rc)

    def post_recv(self, RecvWR wr not None, RecvWR bad_wr=None):
        """
        Post a receive WR on the WQ.
        :param wr: The work request to post
        :param bad_wr: A RecvWR object to hold the bad WR if it is available in
                       case of a failure
        :return: None
        """
        cdef v.ibv_recv_wr *my_bad_wr
        # In order to provide a pointer to a pointer, use a temporary cdef'ed
        # variable.
        rc = v.ibv_post_wq_recv(self.wq, &wr.recv_wr, &my_bad_wr)
        if rc != 0:
            if (bad_wr):
                memcpy(&bad_wr.recv_wr, my_bad_wr, sizeof(bad_wr.recv_wr))
            raise PyverbsRDMAError('Failed to post recv', rc)

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        """
        Closes the underlying C object of the WQ.
        :return: None
        """
        if self.wq != NULL:
            if self.logger:
                self.logger.debug('Closing WQ')
            close_weakrefs([self.rwq_ind_tables])
            rc = v.ibv_destroy_wq(self.wq)
            if rc != 0:
                raise PyverbsRDMAError('Failed to dealloc WQ', rc)
            self.wq = NULL
            self.context = None
            self.pd = None
            self.cq = None

    @property
    def wqn(self):
        return self.wq.wq_num


cdef class RwqIndTableInitAttr(PyverbsObject):
    def __init__(self, log_ind_tbl_size=5, wqs_list=None, comp_mask=0):
        """
        Initializes a RwqIndTableInitAttr object representing ibv_rwq_ind_table_init_attr struct.
        :param log_ind_tbl_size: Log, base 2, of Indirection table size
        :param wqs_list: List of WQs
        :param comp_mask: Identifies valid fields
        :return: A RwqIndTableInitAttr object
        """
        super().__init__()
        if log_ind_tbl_size <= 0:
            raise PyverbsUserError('Invalid indirection table size. Log size must be > 0')
        if (1 << log_ind_tbl_size) < len(wqs_list):
            raise PyverbsUserError(f'Requested table size ({1 << log_ind_tbl_size}) is smaller '
                                   f'than the number of wqs ({len(wqs_list)})')
        self.attr.log_ind_tbl_size = log_ind_tbl_size
        cdef v.ibv_wq **rwq_ind_table = <v.ibv_wq **>calloc(len(wqs_list), sizeof(v.ibv_wq*))
        if rwq_ind_table == NULL:
            raise MemoryError('Failed to allocate memory for Indirection Table')
        for i in range(len(wqs_list)):
            rwq_ind_table[i] = (<WQ>wqs_list[i]).wq
        self.attr.ind_tbl = rwq_ind_table
        self.wqs_list = wqs_list
        self.attr.comp_mask = comp_mask

    def __dealloc__(self):
        """
        Closes the rwq_ind_tbl init attr.
        :return: None
        """
        free(self.attr.ind_tbl)


cdef class RwqIndTable(PyverbsCM):
    def __init__(self, Context ctx, RwqIndTableInitAttr attr):
        """
        Initializes a RwqIndTable object.
        :param ctx: The context the RWQ IND TBL will be associated with.
        :param attr: RWQ IND TBL initial attributes of type RwqIndTableInitAttr.
        :return: A RwqIndTable object
        """
        super().__init__()
        self.rwq_ind_table = v.ibv_create_rwq_ind_table(ctx.context, &attr.attr)
        if self.rwq_ind_table == NULL:
            raise PyverbsRDMAErrno('Failed to create RwqIndTable')
        self.context = ctx
        ctx.add_ref(self)
        self.wqs = attr.wqs_list
        for wq in self.wqs:
            wq.add_ref(self)
        self.qps = weakref.WeakSet()

    cpdef add_ref(self, obj):
        if isinstance(obj, QP):
            self.qps.add(obj)
        else:
            raise PyverbsError('Unrecognized object type')

    @property
    def wqs(self):
        return self.wqs

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        """
        Closes the underlying C object of the RWQ IND TBL.
        :return: None
        """
        if self.rwq_ind_table != NULL:
            if self.logger:
                self.logger.debug('Closing RWQ IND TBL')
            close_weakrefs([self.qps])
            rc = v.ibv_destroy_rwq_ind_table(self.rwq_ind_table)
            if rc != 0:
                raise PyverbsRDMAError('Failed to dealloc RWQ IND TBL', rc)
            self.rwq_ind_table = NULL
            self.context = None


cdef class RxHashConf(PyverbsObject):
    def __init__(self, rx_hash_function=0, rx_hash_key_len=0,
                 rx_hash_key=None, rx_hash_fields_mask=0):
        """
        Initializes a RxHashConf object representing ibv_rx_hash_conf struct.
        :param rx_hash_function: RX hash function, use enum ibv_rx_hash_function_flags
        :param rx_hash_key_len: RX hash key length
        :param rx_hash_key: RX hash key data
        :param rx_hash_fields_mask: RX fields that should participate in the hashing
        :return: A RxHashConf object
        """
        super().__init__()
        cdef uint8_t *rx_hash_key_c = NULL
        if rx_hash_key:
            if rx_hash_key_len != len(rx_hash_key):
                raise PyverbsUserError('Length of rx_hash_key not equal to rx_hash_key_len')
            self.rx_hash_key = rx_hash_key
        self.rx_hash_conf.rx_hash_function = rx_hash_function
        self.rx_hash_conf.rx_hash_key_len = rx_hash_key_len
        self.rx_hash_conf.rx_hash_fields_mask = rx_hash_fields_mask

    @property
    def rx_hash_function(self):
        return self.rx_hash_conf.rx_hash_function
    @rx_hash_function.setter
    def rx_hash_function(self, val):
        self.rx_hash_conf.rx_hash_function = val

    @property
    def rx_hash_key_len(self):
        return self.rx_hash_conf.rx_hash_key_len
    @rx_hash_key_len.setter
    def rx_hash_key_len(self, val):
        if val <= 0:
            raise PyverbsUserError('Invalid rx_hash_key_len. Must be greater then 0')
        self.rx_hash_conf.rx_hash_key_len = val

    @property
    def rx_hash_fields_mask(self):
        return self.rx_hash_conf.rx_hash_fields_mask
    @rx_hash_fields_mask.setter
    def rx_hash_fields_mask(self, val):
        self.rx_hash_conf.rx_hash_fields_mask = val

    @property
    def rx_hash_key(self):
        return self.rx_hash_conf.rx_hash_fields_mask
    @rx_hash_key.setter
    def rx_hash_key(self, vals_list):
        if self.rx_hash_conf.rx_hash_key != NULL:
            free(self.rx_hash_conf.rx_hash_key)
            self.rx_hash_conf.rx_hash_key = NULL
        cdef uint8_t *rx_hash_key_c = <uint8_t*>calloc(len(vals_list), sizeof(uint8_t))
        if rx_hash_key_c == NULL:
            raise MemoryError('Failed to allocate memory for RX hash key')
        for i in range(len(vals_list)):
            rx_hash_key_c[i] = vals_list[i]
        self.rx_hash_conf.rx_hash_key = rx_hash_key_c
        self.rx_hash_conf.rx_hash_key_len = len(vals_list)

    def __dealloc__(self):
        """
        Frees rx hash key allocated memory.
        :return: None
        """
        free(self.rx_hash_conf.rx_hash_key)
        self.rx_hash_conf.rx_hash_key = NULL



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2022 Nvidia Inc. All rights reserved. See COPYING file

import resource

from pyverbs.providers.mlx5.mlx5dv import Mlx5DevxObj, WqeDataSeg, Mlx5UMEM
from tests.mlx5_base import Mlx5DevxRcResources, Mlx5DevxTrafficBase
import pyverbs.providers.mlx5.mlx5_enums as dve
import tests.cuda_utils as cu
import pyverbs.enums as e

try:
    from cuda import cuda, cudart, nvrtc
    cu.CUDA_FOUND = True
except ImportError:
    cu.CUDA_FOUND = False

GPU_PAGE_SIZE = 1 << 16


@cu.set_mem_io_cuda_methods
class CudaDevxRes(Mlx5DevxRcResources):
    def __init__(self, dev_name, ib_port, gid_index,
                 mr_access=e.IBV_ACCESS_LOCAL_WRITE):
        """
        Initialize DevX resources with CUDA memory allocations.
        :param dev_name: Device name to be used
        :param ib_port: IB port of the device to use
        :param gid_index: Which GID index to use
        :param mr_access: The MR access
        """
        self.mr_access = mr_access
        self.cuda_addr = None
        self.dmabuf_fd = None
        self.umem = None
        self.mkey = None
        self.lkey = None
        self.lkey = None
        super().__init__(dev_name=dev_name, ib_port=ib_port, gid_index=gid_index)

    def init_resources(self):
        self.alloc_cuda_mem()
        super().init_resources()
        self.create_dmabuf_umem()
        self.create_mkey()

    def get_wqe_data_segment(self):
        return WqeDataSeg(self.msg_size, self.lkey, int(self.cuda_addr))

    def alloc_cuda_mem(self):
        """
        Allocates CUDA memory and a DMABUF FD on that memory.
        """
        self.cuda_addr = cu.check_cuda_errors(cuda.cuMemAlloc(GPU_PAGE_SIZE))

        # Sync between memory operations
        attr_value = 1
        cu.check_cuda_errors(cuda.cuPointerSetAttribute(
            attr_value,
            cuda.CUpointer_attribute.CU_POINTER_ATTRIBUTE_SYNC_MEMOPS,
            int(self.cuda_addr)
        ))

        # Memory address and size must be aligned to page size to get a handle
        assert (GPU_PAGE_SIZE % resource.getpagesize() == 0 and
                int(self.cuda_addr) % resource.getpagesize() == 0)
        self.dmabuf_fd = cu.check_cuda_errors(
            cuda.cuMemGetHandleForAddressRange(self.cuda_addr,
                                               GPU_PAGE_SIZE,
                                               cuda.CUmemRangeHandleType.CU_MEM_RANGE_HANDLE_TYPE_DMA_BUF_FD,
                                               0))

    def create_mr(self):
        pass

    def create_dmabuf_umem(self):
        umem_aligment = resource.getpagesize()
        self.umem = Mlx5UMEM(self.ctx, GPU_PAGE_SIZE, 0,
                             umem_aligment, self.mr_access, umem_aligment,
                             dve.MLX5DV_UMEM_MASK_DMABUF, self.dmabuf_fd)

    def create_mkey(self):
        from tests.mlx5_prm_structs import SwMkc, CreateMkeyIn, CreateMkeyOut
        accesses = [e.IBV_ACCESS_LOCAL_WRITE, e.IBV_ACCESS_REMOTE_READ, e.IBV_ACCESS_REMOTE_WRITE]
        lw, rr, rw = (list(map(lambda access: int(self.mr_access & access != 0), accesses)))
        mkey_ctx = SwMkc(lr=1, lw=lw, rr=rr, rw=rw, access_mode_1_0=0x1,
                         start_addr=int(self.cuda_addr),
                         len=GPU_PAGE_SIZE, pd=self.dv_pd.pdn, qpn=0xffffff)
        self.mkey = Mlx5DevxObj(self.ctx, CreateMkeyIn(sw_mkc=mkey_ctx,
                                                       mkey_umem_id=self.umem.umem_id,
                                                       mkey_umem_valid=1),
                                len(CreateMkeyOut()))
        self.lkey = CreateMkeyOut(self.mkey.out_view).mkey_index << 8


@cu.set_init_cuda_methods
class Mlx5GpuDevxRcTrafficTest(Mlx5DevxTrafficBase):
    """
    Test DevX traffic over CUDA memory using DMA BUF and UMEM
    """

    @cu.requires_cuda
    def test_mlx_devx_cuda_send_imm_traffic(self):
        """
        Creates two DevX RC QPs and runs SEND_IMM traffic over CUDA allocated
        memory using UMEM and DMA BUF.
        """
        self.create_players(CudaDevxRes)
        # Send traffic
        self.send_imm_traffic()



#!/usr/bin/env python
# Copyright (c) 2018 Mellanox Technologies, Ltd.  All rights reserved.
# Licensed under BSD (MIT variant) or GPLv2. See COPYING.
"""This tool is used to create installable versions of the static libraries in rdma-core.

This is complicated because rdma-core was not designed with static libraries
in mind and relies on the dynamic linker to hide a variety of internal
details.

The build uses several internal utility libraries across the providers and the
libraries. When building statically these libraries have to become inlined
into the various main libraries. This script figures out which static
libraries should include which internal libraries and inlines them
appropriately.

rdma-core is not careful to use globally unique names throughout all the
libraries and all the providers.  Normally the map file in the dynamic linker
will hide these external symbols. This script does something similar for static
linking by analyzing the libraries and map files then renaming internal
symbols with a globally unique prefix.

This is far too complicated to handle internally with cmake, so we have cmake
produce the nearly completed libraries, then process them here using bintuils,
and finally produce the final installation ready libraries."""

import collections
import subprocess
import argparse
import tempfile
import itertools
import sys
import os
import re

SymVer = collections.namedtuple(
    "SymVer", ["version", "prior_version", "globals", "locals"])

try:
    from tempfile import TemporaryDirectory
except ImportError:
    import shutil
    import tempfile

    # From /usr/lib/python3/dist-packages/setuptools/py31compat.py
    class TemporaryDirectory(object):
        def __init__(self):
            self.name = None
            self.name = tempfile.mkdtemp()

        def __enter__(self):
            return self.name

        def __exit__(self, exctype, excvalue, exctrace):
            try:
                shutil.rmtree(self.name, True)
            except OSError:
                pass
            self.name = None


try:
    from subprocess import check_output
except ImportError:
    # From /usr/lib/python2.7/subprocess.py
    def check_output(*popenargs, **kwargs):
        if 'stdout' in kwargs:
            raise ValueError(
                'stdout argument not allowed, it will be overridden.')
        process = subprocess.Popen(
            stdout=subprocess.PIPE, *popenargs, **kwargs)
        output, unused_err = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            raise CalledProcessError(retcode, cmd, output=output)
        return output

    subprocess.check_output = check_output


def parse_stanza(version, prior_version, lines):
    gbl = []
    local = []
    cur = None

    cur = 0
    for I in re.finditer(
            r"\s*(?:(global:)|(local:)(\s*\*\s*;)|(?:(\w+)\s*;))",
            lines,
            flags=re.DOTALL | re.MULTILINE):
        if I.group(1):  # global
            lst = gbl
        if I.group(2):  # local
            lst = local
        if I.group(3):  # wildcard
            lst.append("*")
            assert (cur is not gbl)
        if I.group(4):  # symbol name
            lst.append(I.group(4))

        assert cur == I.start()
        cur = I.end()
    assert cur == len(lines)

    return SymVer(version or "", prior_version or "", gbl, local)


def load_map(fn):
    """This is a lame regex based parser for GNU linker map files. It asserts if
    the map file is invalid. It returns a list of the global symbols"""
    with open(fn, "rt") as F:
        lines = F.read()
    p = re.compile(r"/\*.*?\*/", flags=re.DOTALL)
    lines = re.sub(p, "", lines)
    lines = lines.strip()

    # Extract each stanza
    res = []
    cur = 0
    for I in re.finditer(
            r"\s*(?:(\S+)\s+)?{(.*?)\s*}(\s*\S+)?\s*;",
            lines,
            flags=re.DOTALL | re.MULTILINE):
        assert cur == I.start()
        res.append(parse_stanza(I.group(1), I.group(3), I.group(2)))
        cur = I.end()
    assert cur == len(lines)

    return res


class Lib(object):
    def __init__(self, libfn, tmpdir):
        self.libfn = os.path.basename(libfn)
        self.objdir = os.path.join(tmpdir, self.libfn)
        self.final_objdir = os.path.join(tmpdir, "r-" + self.libfn)
        self.final_lib = os.path.join(os.path.dirname(libfn), "..", self.libfn)
        self.needs = set()
        self.needed = set()

        os.makedirs(self.objdir)
        os.makedirs(self.final_objdir)

        subprocess.check_call([args.ar, "x", libfn], cwd=self.objdir)
        self.objects = [I for I in os.listdir(self.objdir)]
        self.get_syms()

    def get_syms(self):
        """Read the definedsymbols from each object file"""
        self.syms = set()
        self.needed_syms = set()
        for I in self.objects:
            I = os.path.join(self.objdir, I)
            syms = subprocess.check_output([args.nm, "--defined-only", I])
            for ln in syms.decode().splitlines():
                ln = ln.split()
                if ln[1].isupper():
                    self.syms.add(ln[2])

            syms = subprocess.check_output([args.nm, "--undefined-only", I])
            for ln in syms.decode().splitlines():
                ln = ln.split()
                if ln[0].isupper():
                    if not ln[1].startswith("verbs_provider_"):
                        self.needed_syms.add(ln[1])

    def rename_syms(self, rename_fn):
        """Invoke objcopy on all the objects to rename their symbols"""
        for I in self.objects:
            subprocess.check_call([
                args.objcopy,
                "--redefine-syms=%s" % (rename_fn),
                os.path.join(self.objdir, I),
                os.path.join(self.final_objdir, I)
            ])

    def incorporate_internal(self, internal_libs):
        """If this library requires an internal library then we want to inline it into
        this lib when we reconstruct it."""
        for lib in self.needs.intersection(internal_libs):
            self.objects.extend(
                os.path.join(lib.final_objdir, I) for I in lib.objects)

    def finalize(self):
        """Write out the now modified library"""
        try:
            os.unlink(self.final_lib)
        except OSError:
            pass
        subprocess.check_call(
            [args.ar, "qsc", self.final_lib] +
            [os.path.join(self.final_objdir, I) for I in self.objects])


def compute_graph(libs):
    """Look at the symbols each library provides vs the symbols each library needs
    and organize the libraries into a graph."""
    for a, b in itertools.permutations(libs, 2):
        if not a.syms.isdisjoint(b.needed_syms):
            b.needs.add(a)
            a.needed.add(b)

    # Use transitivity to prune the needs list
    def prune(cur_lib, to_prune):
        for I in cur_lib.needed:
            I.needs.discard(to_prune)
            to_prune.needed.discard(I)
            prune(I, to_prune)

    for cur_lib in libs:
        for I in list(cur_lib.needed):
            prune(I, cur_lib)


parser = argparse.ArgumentParser(
    description='Generate static libraries for distribution')
parser.add_argument(
    "--map",
    dest="maps",
    action="append",
    help="List of map files defining all the public symbols",
    default=[])
parser.add_argument(
    "--lib", dest="libs", action="append", help="The input static libraries")
parser.add_argument(
    "--internal_lib",
    dest="internal_libs",
    action="append",
    help=
    "The internal static libraries, these will be merged into other libraries")
parser.add_argument(
    "--version", action="store", help="Package version number", required=True)
parser.add_argument("--ar", action="store", help="ar tool", required=True)
parser.add_argument("--nm", action="store", help="nm tool", required=True)
parser.add_argument(
    "--objcopy", action="store", help="objcopy tool", required=True)
args = parser.parse_args()

global_syms = set()
for fn in sorted(set(args.maps)):
    for I in load_map(fn):
        # Private symbols in libibverbs are also mangled for maximum safety.
        if "PRIVATE" not in I.version:
            global_syms.update(I.globals)

with TemporaryDirectory() as tmpdir:
    libs = set(Lib(fn, tmpdir) for fn in args.libs)
    internal_libs = set(Lib(fn, tmpdir) for fn in args.internal_libs)
    all_libs = libs | internal_libs

    all_syms = set()
    for I in all_libs:
        all_syms.update(I.syms)
    compute_graph(all_libs)

    # To support the ibv_static_providers() machinery these are made global
    # too, even though they are not in map files. We only want to expose them
    # for the static linking case.
    global_syms.add("ibv_static_providers")
    for I in all_syms:
        if I.startswith("verbs_provider_"):
            global_syms.add(I)

    # Generate a redefine file for objcopy that will sanitize the internal names
    prefix = re.sub(r"\W", "_", args.version)
    redefine_fn = os.path.join(tmpdir, "redefine")
    with open(redefine_fn, "wt") as F:
        for I in sorted(all_syms - global_syms):
            F.write("%s rdmacore%s_%s\n" % (I, prefix, I))

    for I in all_libs:
        I.rename_syms(redefine_fn)

    for I in libs:
        I.incorporate_internal(internal_libs)
        I.finalize()



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019 Mellanox Technologies, Inc. All rights reserved. See COPYING file

#cython: language_level=3

from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t, uintptr_t
from posix.types cimport off_t
from libcpp cimport bool
cimport libc.stdio as s

from pyverbs.providers.mlx5.mlx5_enums cimport *
cimport pyverbs.libibverbs as v


cdef extern from 'infiniband/mlx5dv.h':

    cdef struct mlx5dv_context_attr:
        unsigned int    flags
        unsigned long   comp_mask

    cdef struct mlx5dv_cqe_comp_caps:
        unsigned int    max_num
        unsigned int    supported_format

    cdef struct mlx5dv_sw_parsing_caps:
        unsigned int    sw_parsing_offloads
        unsigned int    supported_qpts

    cdef struct mlx5dv_striding_rq_caps:
        unsigned int    min_single_stride_log_num_of_bytes
        unsigned int    max_single_stride_log_num_of_bytes
        unsigned int    min_single_wqe_log_num_of_strides
        unsigned int    max_single_wqe_log_num_of_strides
        unsigned int    supported_qpts

    cdef struct mlx5dv_dci_streams_caps:
        uint8_t    max_log_num_concurent
        uint8_t    max_log_num_errored

    cdef struct mlx5dv_crypto_caps:
        uint16_t failed_selftests
        uint8_t crypto_engines
        uint8_t wrapped_import_method
        uint8_t log_max_num_deks
        uint32_t flags

    cdef struct mlx5dv_context:
        unsigned char           version
        unsigned long           flags
        unsigned long           comp_mask
        mlx5dv_cqe_comp_caps    cqe_comp_caps
        mlx5dv_sw_parsing_caps  sw_parsing_caps
        mlx5dv_striding_rq_caps striding_rq_caps
        mlx5dv_dci_streams_caps dci_streams_caps
        unsigned int            tunnel_offloads_caps
        unsigned int            max_dynamic_bfregs
        unsigned long           max_clock_info_update_nsec
        unsigned int            flow_action_flags
        unsigned int            dc_odp_caps
        uint8_t                 num_lag_ports
        mlx5dv_crypto_caps      crypto_caps
        size_t                  max_wr_memcpy_length
        uint64_t                max_dc_rd_atom
        uint64_t                max_dc_init_rd_atom


    cdef struct mlx5dv_dci_streams:
        uint8_t       log_num_concurent
        uint8_t       log_num_errored

    cdef struct mlx5dv_dc_init_attr:
        mlx5dv_dc_type      dc_type
        unsigned long       dct_access_key
        mlx5dv_dci_streams  dci_streams

    cdef struct mlx5dv_qp_init_attr:
        unsigned long       comp_mask
        unsigned int        create_flags
        mlx5dv_dc_init_attr dc_init_attr
        unsigned long       send_ops_flags

    cdef struct mlx5dv_cq_init_attr:
        unsigned long   comp_mask
        unsigned char   cqe_comp_res_format
        unsigned int    flags
        unsigned short  cqe_size

    cdef struct mlx5dv_var:
        uint32_t    page_id
        uint32_t    length
        long        mmap_off
        uint64_t    comp_mask

    cdef struct mlx5dv_pp:
        uint16_t index

    cdef struct mlx5dv_devx_uar:
        void        *reg_addr;
        void        *base_addr;
        uint32_t    page_id;
        long        mmap_off;
        uint64_t    comp_mask;

    cdef struct mlx5dv_qp_ex:
        uint64_t comp_mask

    cdef struct mlx5dv_sched_node

    cdef struct mlx5dv_sched_leaf

    cdef struct mlx5dv_sched_attr:
        mlx5dv_sched_node *parent;
        uint32_t flags;
        uint32_t bw_share;
        uint32_t max_avg_bw;
        uint64_t comp_mask;

    cdef struct mlx5dv_reg:
        uint32_t value;
        uint32_t mask;

    cdef struct mlx5dv_port:
        uint64_t flags
        uint16_t vport
        uint16_t vport_vhca_id
        uint16_t esw_owner_vhca_id
        uint64_t vport_steering_icm_rx
        uint64_t vport_steering_icm_tx
        mlx5dv_reg reg_c0

    cdef struct mlx5dv_flow_match_parameters:
        size_t   match_sz;
        uint64_t *match_buf;

    cdef struct mlx5dv_flow_matcher_attr:
        v.ibv_flow_attr_type         type;
        uint32_t                     flags;
        uint16_t                     priority;
        uint8_t                      match_criteria_enable;
        mlx5dv_flow_match_parameters *match_mask;
        uint64_t                     comp_mask;
        mlx5_ib_uapi_flow_table_type ft_type;

    cdef struct mlx5dv_flow_matcher

    cdef struct mlx5dv_devx_obj

    cdef struct mlx5dv_flow_action_attr:
        mlx5dv_flow_action_type type
        v.ibv_qp                *qp
        v.ibv_flow_action       *action
        unsigned int            tag_value
        mlx5dv_devx_obj         *obj

    cdef struct mlx5dv_dr_domain

    cdef struct mlx5dv_dr_table

    cdef struct mlx5dv_dr_matcher

    cdef struct mlx5dv_dr_matcher_layout:
        uint32_t flags
        uint32_t log_num_of_rules_hint

    cdef struct mlx5dv_dr_action

    cdef struct mlx5dv_dr_rule

    cdef struct mlx5dv_dr_action_dest_reformat:
        mlx5dv_dr_action *reformat
        mlx5dv_dr_action *dest

    cdef struct mlx5dv_dr_action_dest_attr:
        mlx5dv_dr_action_dest_type type
        mlx5dv_dr_action *dest
        mlx5dv_dr_action_dest_reformat *dest_reformat

    cdef struct mlx5dv_dr_flow_sampler_attr:
        uint32_t        sample_ratio
        mlx5dv_dr_table    *default_next_table
        uint32_t        num_sample_actions
        mlx5dv_dr_action    **sample_actions
        uint64_t        action

    cdef struct mlx5dv_dr_flow_meter_attr:
        mlx5dv_dr_table *next_table
        uint8_t         active
        uint8_t         reg_c_index
        size_t          flow_meter_parameter_sz
        void            *flow_meter_parameter

    cdef struct mlx5dv_clock_info:
        pass

    cdef struct mlx5dv_mkey_init_attr:
        v.ibv_pd *pd
        uint32_t create_flags
        uint16_t max_entries

    cdef struct mlx5dv_mkey:
        uint32_t lkey
        uint32_t rkey

    cdef struct mlx5dv_mr_interleaved:
        uint64_t addr
        uint32_t bytes_count
        uint32_t bytes_skip
        uint32_t lkey

    cdef struct mlx5dv_mkey_conf_attr:
        uint32_t conf_flags
        uint64_t comp_mask

    cdef struct mlx5dv_sig_crc:
        mlx5dv_sig_crc_type type
        uint64_t seed

    cdef struct mlx5dv_sig_t10dif:
        mlx5dv_sig_t10dif_bg_type bg_type
        uint16_t bg
        uint16_t app_tag
        uint32_t ref_tag
        uint16_t flags

    cdef union sig:
        mlx5dv_sig_t10dif *dif
        mlx5dv_sig_crc *crc

    cdef struct mlx5dv_sig_block_domain:
        mlx5dv_sig_type sig_type
        sig sig
        mlx5dv_block_size block_size
        uint64_t comp_mask

    cdef struct mlx5dv_sig_block_attr:
        mlx5dv_sig_block_domain *mem
        mlx5dv_sig_block_domain *wire
        uint32_t flags
        uint8_t check_mask
        uint8_t copy_mask
        uint64_t comp_mask

    cdef struct mlx5dv_sig_err:
        uint64_t actual_value
        uint64_t expected_value
        uint64_t offset

    cdef union err:
        mlx5dv_sig_err sig

    cdef struct mlx5dv_mkey_err:
        mlx5dv_mkey_err_type err_type
        err err

    cdef struct mlx5_wqe_data_seg:
        uint32_t    byte_count
        uint32_t    lkey
        uint64_t    addr

    cdef struct mlx5_wqe_ctrl_seg:
        uint32_t    opmod_idx_opcode
        uint32_t    qpn_ds
        uint8_t     signature
        uint8_t     fm_ce_se
        uint32_t    imm

    cdef struct mlx5dv_devx_umem:
        uint32_t umem_id;

    cdef struct mlx5dv_devx_umem_in:
        void        *addr
        size_t      size
        uint32_t    access
        uint64_t    pgsz_bitmap
        uint64_t    comp_mask
        int         dmabuf_fd

    cdef struct mlx5dv_vfio_context_attr:
        const char  *pci_name
        uint32_t    flags
        uint64_t    comp_mask

    cdef struct mlx5dv_devx_msi_vector:
        int         vector
        int         fd

    cdef struct mlx5dv_devx_eq:
        void        *vaddr

    cdef struct mlx5dv_pd:
        uint32_t    pdn
        uint64_t    comp_mask

    cdef struct mlx5dv_cq:
        void        *buf
        uint32_t    *dbrec
        uint32_t    cqe_cnt
        uint32_t    cqe_size
        void        *cq_uar
        uint32_t    cqn
        uint64_t    comp_mask

    cdef struct mlx5dv_qp:
        uint64_t    comp_mask
        off_t       uar_mmap_offset
        uint32_t    tirn
        uint32_t    tisn
        uint32_t    rqn
        uint32_t    sqn

    cdef struct mlx5dv_srq:
        uint32_t    stride
        uint32_t    head
        uint32_t    tail
        uint64_t    comp_mask
        uint32_t    srqn

    cdef struct pd:
        v.ibv_pd    *in_ "in"
        mlx5dv_pd   *out

    cdef struct cq:
        v.ibv_cq    *in_ "in"
        mlx5dv_cq   *out

    cdef struct qp:
        v.ibv_qp    *in_ "in"
        mlx5dv_qp   *out

    cdef struct srq:
        v.ibv_srq   *in_ "in"
        mlx5dv_srq  *out

    cdef struct mlx5dv_obj:
        pd  pd
        cq  cq
        qp  qp
        srq srq

    cdef struct mlx5_cqe64:
        uint16_t    wqe_id
        uint32_t    imm_inval_pkey
        uint32_t    byte_cnt
        uint64_t    timestamp
        uint16_t    wqe_counter
        uint8_t     signature
        uint8_t     op_own


    void mlx5dv_set_ctrl_seg(mlx5_wqe_ctrl_seg *seg, uint16_t pi, uint8_t opcode,
                             uint8_t opmod, uint32_t qp_num, uint8_t fm_ce_se,
                             uint8_t ds, uint8_t signature, uint32_t imm)
    void mlx5dv_set_data_seg(mlx5_wqe_data_seg *seg, uint32_t length,
                             uint32_t lkey, uintptr_t address)
    uint8_t mlx5dv_get_cqe_owner(mlx5_cqe64 *cqe)
    void mlx5dv_set_cqe_owner(mlx5_cqe64 *cqe, uint8_t val)
    uint8_t mlx5dv_get_cqe_se(mlx5_cqe64 *cqe)
    uint8_t mlx5dv_get_cqe_format(mlx5_cqe64 *cqe)
    uint8_t mlx5dv_get_cqe_opcode(mlx5_cqe64 *cqe)

    cdef struct mlx5dv_dek:
        pass

    cdef struct mlx5dv_crypto_login_obj:
        pass

    cdef struct mlx5dv_crypto_login_attr:
        uint32_t credential_id
        uint32_t import_kek_id
        char *credential
        uint64_t comp_mask

    cdef struct mlx5dv_crypto_login_attr_ex:
        uint32_t credential_id
        uint32_t import_kek_id
        const void *credential
        size_t credential_len
        uint64_t comp_mask

    cdef struct mlx5dv_crypto_login_query_attr:
        mlx5dv_crypto_login_state state
        uint64_t comp_mask

    cdef struct mlx5dv_crypto_attr:
        mlx5dv_crypto_standard crypto_standard
        bool encrypt_on_tx
        mlx5dv_signature_crypto_order signature_crypto_order
        mlx5dv_block_size data_unit_size
        char *initial_tweak
        mlx5dv_dek *dek
        char *keytag
        uint64_t comp_mask

    cdef struct mlx5dv_dek_init_attr:
        mlx5dv_crypto_key_size key_size
        bool has_keytag
        mlx5dv_crypto_key_purpose key_purpose
        v.ibv_pd *pd
        char *opaque
        char *key
        uint64_t comp_mask
        mlx5dv_crypto_login_obj *crypto_login

    cdef struct mlx5dv_dek_attr:
        mlx5dv_dek_state state
        char *opaque
        uint64_t comp_mask

    bool mlx5dv_is_supported(v.ibv_device *device)
    v.ibv_context* mlx5dv_open_device(v.ibv_device *device,
                                      mlx5dv_context_attr *attr)
    int mlx5dv_query_device(v.ibv_context *ctx, mlx5dv_context *attrs_out)

    v.ibv_qp *mlx5dv_create_qp(v.ibv_context *context,
                               v.ibv_qp_init_attr_ex *qp_attr,
                               mlx5dv_qp_init_attr *mlx5_qp_attr)
    int mlx5dv_query_qp_lag_port(v.ibv_qp *qp, uint8_t *port_num,
                                 uint8_t *active_port_num)
    int mlx5dv_modify_qp_lag_port(v.ibv_qp *qp, uint8_t port_num)
    int mlx5dv_modify_qp_udp_sport(v.ibv_qp *qp, uint16_t udp_sport)
    int mlx5dv_dci_stream_id_reset(v.ibv_qp *qp, uint16_t stream_id)
    v.ibv_cq_ex *mlx5dv_create_cq(v.ibv_context *context,
                                  v.ibv_cq_init_attr_ex *cq_attr,
                                  mlx5dv_cq_init_attr *mlx5_cq_attr)
    void mlx5dv_wr_raw_wqe(mlx5dv_qp_ex *mqp_ex, const void *wqe)
    mlx5dv_var *mlx5dv_alloc_var(v.ibv_context *context, uint32_t flags)
    void mlx5dv_free_var(mlx5dv_var *dv_var)
    mlx5dv_pp *mlx5dv_pp_alloc(v.ibv_context *context, size_t pp_context_sz,
                               const void *pp_context, uint32_t flags)
    void mlx5dv_pp_free(mlx5dv_pp *pp)
    void mlx5dv_wr_set_dc_addr(mlx5dv_qp_ex *mqp, v.ibv_ah *ah,
                               uint32_t remote_dctn, uint64_t remote_dc_key)
    void mlx5dv_wr_set_dc_addr_stream(mlx5dv_qp_ex *mqp, v.ibv_ah *ah,
                                      uint32_t remote_dctn, uint64_t remote_dc_key,
                                      uint16_t stream_id)
    void mlx5dv_wr_mr_interleaved(mlx5dv_qp_ex *mqp, mlx5dv_mkey *mkey,
                                  uint32_t access_flags, uint32_t repeat_count,
                                  uint16_t num_interleaved, mlx5dv_mr_interleaved *data)
    void mlx5dv_wr_mr_list(mlx5dv_qp_ex *mqp, mlx5dv_mkey *mkey,
                           uint32_t access_flags, uint16_t num_sge, v.ibv_sge *sge)
    void mlx5dv_wr_memcpy(mlx5dv_qp_ex *mqp, uint32_t dest_lkey, uint64_t dest_addr,
                          uint32_t src_lkey, uint64_t src_addr, uint64_t length)
    mlx5dv_mkey *mlx5dv_create_mkey(mlx5dv_mkey_init_attr *mkey_init_attr)
    int mlx5dv_destroy_mkey(mlx5dv_mkey *mkey)
    mlx5dv_qp_ex *mlx5dv_qp_ex_from_ibv_qp_ex(v.ibv_qp_ex *qp_ex)
    mlx5dv_sched_node *mlx5dv_sched_node_create(v.ibv_context *context,
                                                mlx5dv_sched_attr *sched_attr)
    mlx5dv_sched_leaf *mlx5dv_sched_leaf_create(v.ibv_context *context,
                                                mlx5dv_sched_attr *sched_attr)
    int mlx5dv_sched_node_modify(mlx5dv_sched_node *node,
                                 mlx5dv_sched_attr *sched_attr)
    int mlx5dv_sched_leaf_modify(mlx5dv_sched_leaf *leaf,
                                 mlx5dv_sched_attr *sched_attr)
    int mlx5dv_sched_node_destroy(mlx5dv_sched_node *node)
    int mlx5dv_sched_leaf_destroy(mlx5dv_sched_leaf *leaf)
    int mlx5dv_modify_qp_sched_elem(v.ibv_qp *qp, mlx5dv_sched_leaf *requestor,
                                    mlx5dv_sched_leaf *responder)
    int mlx5dv_reserved_qpn_alloc(v.ibv_context *context, uint32_t *qpn)
    int mlx5dv_reserved_qpn_dealloc(v.ibv_context *context, uint32_t qpn)
    void *mlx5dv_dm_map_op_addr(v.ibv_dm *dm, uint8_t op)
    int mlx5dv_query_port(v.ibv_context *context, uint32_t port_num, mlx5dv_port *port)
    mlx5dv_flow_matcher *mlx5dv_create_flow_matcher(v.ibv_context *context,
                                                    mlx5dv_flow_matcher_attr *matcher_attr)
    int mlx5dv_destroy_flow_matcher(mlx5dv_flow_matcher *matcher)
    v.ibv_flow *mlx5dv_create_flow(mlx5dv_flow_matcher *matcher,
                                   mlx5dv_flow_match_parameters *match_value,
                                   size_t num_actions,
                                   mlx5dv_flow_action_attr actions_attr[])
    v.ibv_flow_action *mlx5dv_create_flow_action_packet_reformat(v.ibv_context *context,
                                                                 size_t data_sz,
                                                                 void *data,
                                                                 unsigned char reformat_type,
                                                                 unsigned char ft_type)

    # Direct rules verbs
    mlx5dv_dr_domain *mlx5dv_dr_domain_create(v.ibv_context *ctx, mlx5dv_dr_domain_type type)
    int mlx5dv_dr_domain_sync(mlx5dv_dr_domain *domain, uint32_t flags)
    int mlx5dv_dump_dr_domain(s.FILE *fout, mlx5dv_dr_domain *domain)
    int mlx5dv_dr_domain_destroy(mlx5dv_dr_domain *dmn)
    mlx5dv_dr_table *mlx5dv_dr_table_create(mlx5dv_dr_domain *dmn, uint32_t level)
    int mlx5dv_dr_table_destroy(mlx5dv_dr_table *tbl)
    mlx5dv_dr_matcher *mlx5dv_dr_matcher_create(mlx5dv_dr_table *table,
                                                uint16_t priority,
                                                uint8_t match_criteria_enable,
                                                mlx5dv_flow_match_parameters *mask)
    int mlx5dv_dr_matcher_set_layout(mlx5dv_dr_matcher *matcher, mlx5dv_dr_matcher_layout *layout)
    int mlx5dv_dr_matcher_destroy(mlx5dv_dr_matcher *matcher)
    mlx5dv_dr_action *mlx5dv_dr_action_create_dest_ibv_qp(v.ibv_qp *ibqp)
    mlx5dv_dr_action *mlx5dv_dr_action_create_tag(uint32_t tag_value)
    mlx5dv_dr_action *mlx5dv_dr_action_create_dest_table(mlx5dv_dr_table *tbl)
    mlx5dv_dr_action *mlx5dv_dr_action_create_pop_vlan()
    mlx5dv_dr_action *mlx5dv_dr_action_create_push_vlan(mlx5dv_dr_domain *dmn,
                                                        uint32_t vlan_hdr)
    mlx5dv_dr_action *mlx5dv_dr_action_create_dest_array(
            mlx5dv_dr_domain *domain, size_t num_dest,
            mlx5dv_dr_action_dest_attr *dests[])
    int mlx5dv_dr_action_destroy(mlx5dv_dr_action *action)
    mlx5dv_dr_rule *mlx5dv_dr_rule_create(mlx5dv_dr_matcher *matcher,
                                          mlx5dv_flow_match_parameters *value,
                                          size_t num_actions,
                                          mlx5dv_dr_action *actions[])
    mlx5dv_dr_action *mlx5dv_dr_action_create_modify_header(mlx5dv_dr_domain *dmn, uint32_t flags,
                                                            size_t actions_sz, uint64_t actions[])
    mlx5dv_dr_action *mlx5dv_dr_action_create_flow_counter(mlx5dv_devx_obj *devx_obj,
                                                           uint32_t offset)
    mlx5dv_dr_action *mlx5dv_dr_action_create_drop()
    mlx5dv_dr_action *mlx5dv_dr_action_create_default_miss()
    mlx5dv_dr_action *mlx5dv_dr_action_create_dest_vport(mlx5dv_dr_domain *dmn,
                                                         uint32_t vport)
    mlx5dv_dr_action *mlx5dv_dr_action_create_dest_ib_port(mlx5dv_dr_domain *dmn,
                                                           uint32_t ib_port)
    mlx5dv_dr_action *mlx5dv_dr_action_create_packet_reformat(mlx5dv_dr_domain *domain,
                                                              uint32_t flags,
                                                              unsigned char reformat_type,
                                                              size_t data_sz, void *data)
    int mlx5dv_dr_rule_destroy(mlx5dv_dr_rule *rule)
    void mlx5dv_dr_domain_allow_duplicate_rules(mlx5dv_dr_domain *dmn, bool allow)

    uint64_t mlx5dv_ts_to_ns(mlx5dv_clock_info *clock_info,
                             uint64_t device_timestamp)
    int mlx5dv_get_clock_info(v.ibv_context *ctx_in, mlx5dv_clock_info *clock_info)
    int mlx5dv_map_ah_to_qp(v.ibv_ah *ah, uint32_t qp_num)
    v.ibv_device **mlx5dv_get_vfio_device_list(mlx5dv_vfio_context_attr *attr)
    int mlx5dv_vfio_get_events_fd(v.ibv_context *ibctx)
    int mlx5dv_vfio_process_events(v.ibv_context *context)
    mlx5dv_dr_action *mlx5dv_dr_action_create_dest_devx_tir(mlx5dv_devx_obj *devx_obj)
    mlx5dv_dr_action *mlx5dv_dr_action_create_flow_sampler(mlx5dv_dr_flow_sampler_attr *attr)
    mlx5dv_dr_action *mlx5dv_dr_action_create_flow_meter(mlx5dv_dr_flow_meter_attr *attr)
    int mlx5dv_dr_action_modify_flow_meter(mlx5dv_dr_action *action,
                                           mlx5dv_dr_flow_meter_attr *attr,
                                           uint64_t modify_field_select)

    # DevX APIs
    mlx5dv_devx_uar *mlx5dv_devx_alloc_uar(v.ibv_context *context, uint32_t flags)
    void mlx5dv_devx_free_uar(mlx5dv_devx_uar *devx_uar)
    int mlx5dv_devx_general_cmd(v.ibv_context *context, const void *in_,
                                size_t inlen, void *out, size_t outlen)
    mlx5dv_devx_umem *mlx5dv_devx_umem_reg(v.ibv_context *ctx, void *addr,
                                           size_t size, unsigned long access)
    mlx5dv_devx_umem *mlx5dv_devx_umem_reg_ex(v.ibv_context *ctx,
                                              mlx5dv_devx_umem_in *umem_in)
    int mlx5dv_devx_umem_dereg(mlx5dv_devx_umem *umem)
    int mlx5dv_devx_query_eqn(v.ibv_context *context, uint32_t vector, uint32_t *eqn)
    mlx5dv_devx_obj *mlx5dv_devx_obj_create(v.ibv_context *context, const void *_in,
                                            size_t inlen, void *out, size_t outlen)
    int mlx5dv_devx_obj_query(mlx5dv_devx_obj *obj, const void *in_,
                              size_t inlen, void *out, size_t outlen)
    int mlx5dv_devx_obj_modify(mlx5dv_devx_obj *obj, const void *in_,
                               size_t inlen, void *out, size_t outlen)
    int mlx5dv_devx_obj_destroy(mlx5dv_devx_obj *obj)
    int mlx5dv_init_obj(mlx5dv_obj *obj, uint64_t obj_type)
    mlx5dv_devx_msi_vector *mlx5dv_devx_alloc_msi_vector(v.ibv_context *ibctx)
    int mlx5dv_devx_free_msi_vector(mlx5dv_devx_msi_vector *msi)
    mlx5dv_devx_eq *mlx5dv_devx_create_eq(v.ibv_context *context, const void *_in,
                                          size_t inlen, void *out, size_t outlen)
    int mlx5dv_devx_destroy_eq(mlx5dv_devx_eq *eq)

    # Mkey setters
    void mlx5dv_wr_mkey_configure(mlx5dv_qp_ex *mqp, mlx5dv_mkey *mkey,
                                 int num_setters, mlx5dv_mkey_conf_attr *attr)
    void mlx5dv_wr_set_mkey_access_flags(mlx5dv_qp_ex *mqp, uint32_t access_flags)
    void mlx5dv_wr_set_mkey_layout_list(mlx5dv_qp_ex *mqp, uint16_t num_sges, v.ibv_sge *sge)
    void mlx5dv_wr_set_mkey_layout_interleaved(mlx5dv_qp_ex *mqp, uint32_t repeat_count,
                                               uint16_t num_interleaved,
                                               mlx5dv_mr_interleaved *data)
    void mlx5dv_wr_set_mkey_sig_block(mlx5dv_qp_ex *mqp, mlx5dv_sig_block_attr *attr)
    int mlx5dv_mkey_check(mlx5dv_mkey *mkey, mlx5dv_mkey_err *err_info)
    int mlx5dv_qp_cancel_posted_send_wrs(mlx5dv_qp_ex *mqp, uint64_t wr_id)
    void mlx5dv_wr_set_mkey_crypto(mlx5dv_qp_ex *mqp, mlx5dv_crypto_attr *attr)

    # Crypto APIs
    int mlx5dv_crypto_login(v.ibv_context *context, mlx5dv_crypto_login_attr *login_attr)
    int mlx5dv_crypto_login_query_state(v.ibv_context *context, mlx5dv_crypto_login_state *state)
    int mlx5dv_crypto_logout(v.ibv_context *context)
    mlx5dv_dek *mlx5dv_dek_create(v.ibv_context *context, mlx5dv_dek_init_attr *init_attr)
    int mlx5dv_dek_query(mlx5dv_dek *dek, mlx5dv_dek_attr *attr)
    int mlx5dv_dek_destroy(mlx5dv_dek *dek)
    mlx5dv_crypto_login_obj *mlx5dv_crypto_login_create(v.ibv_context *context,
                                                        mlx5dv_crypto_login_attr_ex *login_attr)
    int mlx5dv_crypto_login_query(mlx5dv_crypto_login_obj *crypto_login,
                                  mlx5dv_crypto_login_query_attr *query_attr)
    int mlx5dv_crypto_login_destroy(mlx5dv_crypto_login_obj *crypto_login)



import unittest

from pyverbs.pyverbs_error import PyverbsRDMAError
from tests.mlx5_base import Mlx5PyverbsAPITestCase
from pyverbs.qp import QP, QPInitAttr, QPCap
from pyverbs.srq import SRQ, SrqInitAttr
from pyverbs.pd import PD
from pyverbs.cq import CQ
import tests.utils as u


def huge_pages_supported():
    try:
        u.huge_pages_supported()
    except unittest.SkipTest:
        return False
    return True


class ResourcesOnHugePageTest(Mlx5PyverbsAPITestCase):
    def create_cq(self):
        return CQ(self.ctx, 100, None, None, 0)

    def create_qp(self):
        with PD(self.ctx) as pd:
            with self.create_cq() as cq:
                attr = QPInitAttr(scq=cq, rcq=cq, cap=QPCap(max_recv_wr=100,
                                                            max_send_wr=100))
                QP(pd, attr)

    def create_srq(self):
        with PD(self.ctx) as pd:
            SRQ(pd, SrqInitAttr())

    def set_env_alloc_type(self, alloc_type):
        self.set_env_variable('MLX_CQ_ALLOC_TYPE', alloc_type)
        self.set_env_variable('MLX_QP_ALLOC_TYPE', alloc_type)
        self.set_env_variable('MLX_SRQ_ALLOC_TYPE', alloc_type)

    def create_objects(self):
        self.create_cq()
        self.create_qp()
        self.create_srq()

    def test_prefer_obj_on_huge(self):
        """
        Test PREFER_HUGE allocation type for srq cq and qp.
        """
        self.set_env_alloc_type('PREFER_HUGE')
        self.create_objects()

    def test_obj_on_huge(self):
        """
        Test HUGE allocation type for srq cq and qp.
        If there are huge pages in the system expect to success,
        else expect to fail.
        """
        self.set_env_alloc_type('HUGE')
        if huge_pages_supported() and u.is_root():
            self.create_objects()
        else:
            with self.assertRaises(PyverbsRDMAError):
                self.create_objects()



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2021 Nvidia, Inc. All rights reserved. See COPYING file

#cython: language_level=3

from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.string cimport strcpy
import weakref

from pyverbs.pyverbs_error import PyverbsRDMAError
cimport pyverbs.providers.mlx5.libmlx5 as dv
from pyverbs.base import PyverbsRDMAErrno
from pyverbs.base cimport close_weakrefs
from pyverbs.device cimport Context
cimport pyverbs.libibverbs as v


cdef class Mlx5VfioAttr(PyverbsObject):
    """
    Mlx5VfioAttr class, represents mlx5dv_vfio_context_attr C struct.
    """
    def __init__(self, pci_name, flags=0, comp_mask=0):
        self.pci_name = pci_name
        self.attr.flags = flags
        self.attr.comp_mask = comp_mask

    def __dealloc__(self):
        if self.attr.pci_name != NULL:
            PyMem_Free(<void*>self.attr.pci_name)
            self.attr.pci_name = NULL

    @property
    def flags(self):
        return self.attr.flags
    @flags.setter
    def flags(self, val):
        self.attr.flags = val

    @property
    def comp_mask(self):
        return self.attr.comp_mask
    @comp_mask.setter
    def comp_mask(self, val):
        self.attr.comp_mask = val

    @property
    def pci_name(self):
        return self.attr.pci_name[:]
    @pci_name.setter
    def pci_name(self, val):
        if self.attr.pci_name != NULL:
            PyMem_Free(<void*>self.attr.pci_name)
        pci_name_bytes = val.encode()
        self.attr.pci_name = <char*>PyMem_Malloc(len(pci_name_bytes))
        strcpy(<char*>self.attr.pci_name, pci_name_bytes)


cdef class Mlx5VfioContext(Mlx5Context):
    """
    Mlx5VfioContext class is used to easily initialize and open a context over
    a mlx5 vfio device.
    It is initialized based on the passed mlx5 vfio attributes (Mlx5VfioAttr),
    by getting the relevant vfio device and opening it (creating a context).
    """
    def __init__(self, Mlx5VfioAttr attr):
        super(Context, self).__init__()
        cdef v.ibv_device **dev_list

        self.name = attr.pci_name
        self.pds = weakref.WeakSet()
        self.devx_umems = weakref.WeakSet()
        self.devx_objs = weakref.WeakSet()
        self.uars = weakref.WeakSet()
        self.devx_eqs = weakref.WeakSet()

        dev_list = dv.mlx5dv_get_vfio_device_list(&attr.attr)
        if dev_list == NULL:
            raise PyverbsRDMAErrno('Failed to get VFIO device list')
        self.device = dev_list[0]
        if self.device == NULL:
            raise PyverbsRDMAError('Failed to get VFIO device')
        try:
            self.context = v.ibv_open_device(self.device)
            if self.context == NULL:
                raise PyverbsRDMAErrno('Failed to open mlx5 VFIO device '
                                       f'({self.device.name.decode()})')
        finally:
            v.ibv_free_device_list(dev_list)

    def get_events_fd(self):
        """
        Gets the file descriptor to manage driver events.
        :return: The file descriptor to be used for managing driver events.
        """
        fd = dv.mlx5dv_vfio_get_events_fd(self.context)
        if fd < 0:
            raise PyverbsRDMAError('Failed to get VFIO events FD', -fd)
        return fd

    def process_events(self):
        """
        Process events on the vfio device.
        This method should run from application thread to maintain device events.
        :return: None
        """
        rc = dv.mlx5dv_vfio_process_events(self.context)
        if rc:
            raise PyverbsRDMAError('VFIO process events failed', rc)

    cpdef close(self):
        if self.context != NULL:
            if self.logger:
                self.logger.debug('Closing Mlx5VfioContext')
            close_weakrefs([self.pds, self.devx_objs, self.devx_umems, self.uars, self.devx_eqs])
            rc = v.ibv_close_device(self.context)
            if rc != 0:
                raise PyverbsRDMAErrno(f'Failed to close device {self.name}')
            self.context = NULL



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright 2020-2023 Amazon.com, Inc. or its affiliates. All rights reserved.

import unittest
import errno

from pyverbs.cq import CQ, CompChannel
from pyverbs.pyverbs_error import PyverbsRDMAError
import pyverbs.enums as e

from tests.efa_base import EfaRDMATestCase
from tests.efa_base import SRDResources
import tests.utils as u


class CqEventsSRD(SRDResources):
    def __init__(self, dev_name, ib_port, gid_index):
        super().__init__(dev_name, ib_port, gid_index, e.IBV_QP_EX_WITH_SEND)

    def create_cq(self):
        self.comp_channel = CompChannel(self.ctx)
        self.cq = CQ(self.ctx, self.num_msgs, None, self.comp_channel)
        self.cq.req_notify()


class CqEventsSRDTestCase(EfaRDMATestCase):
    def setUp(self):
        super().setUp()
        self.iters = 100

    def test_cq_events_srd(self):
        for use_new_send in [False, True]:
            with self.subTest():
                super().create_players(CqEventsSRD)
                u.traffic(**self.traffic_args, new_send=use_new_send)


class QPSRDTestCase(EfaRDMATestCase):
    def setUp(self):
        super().setUp()
        self.iters = 100
        self.server = None
        self.client = None

    def create_players(self, send_ops_flags=0, qp_count=8):
        super().create_players(SRDResources, send_ops_flags=send_ops_flags, qp_count=qp_count)

    def full_sq_bad_flow(self):
        """
        Check post_send while qp's sq is full.
        - Find qp's sq length
        - Fill the qp with work requests until overflow
        """
        qp_idx = 0
        send_op = e.IBV_WR_SEND
        ah = u.get_global_ah(self.client, self.gid_index, self.ib_port)
        qp_attr, _ = self.client.qps[qp_idx].query(e.IBV_QP_CAP)
        max_send_wr = qp_attr.cap.max_send_wr
        with self.assertRaises(PyverbsRDMAError) as ex:
            for _ in range (max_send_wr + 1):
                _, c_sg = u.get_send_elements(self.client, False)
                u.send(self.client, c_sg, send_op, new_send=True, qp_idx=qp_idx, ah=ah)
        self.assertEqual(ex.exception.error_code, errno.ENOMEM)

    def test_qp_ex_srd_send(self):
        self.create_players(e.IBV_QP_EX_WITH_SEND)
        u.traffic(**self.traffic_args, new_send=True, send_op=e.IBV_WR_SEND)

    def test_qp_ex_srd_send_imm(self):
        self.create_players(e.IBV_QP_EX_WITH_SEND_WITH_IMM)
        u.traffic(**self.traffic_args, new_send=True, send_op=e.IBV_WR_SEND_WITH_IMM)

    def test_qp_ex_srd_rdma_read(self):
        self.create_players(e.IBV_QP_EX_WITH_RDMA_READ)
        self.server.mr.write('s' * self.server.msg_size, self.server.msg_size)
        u.rdma_traffic(**self.traffic_args, new_send=True, send_op=e.IBV_WR_RDMA_READ)

    def test_qp_ex_srd_rdma_write(self):
        self.create_players(e.IBV_QP_EX_WITH_RDMA_WRITE)
        u.rdma_traffic(**self.traffic_args, new_send=True, send_op=e.IBV_WR_RDMA_WRITE)

    def test_qp_ex_srd_rdma_write_with_imm(self):
        self.create_players(e.IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM)
        u.traffic(**self.traffic_args, new_send=True, send_op=e.IBV_WR_RDMA_WRITE_WITH_IMM)

    def test_qp_ex_srd_old_send(self):
        self.create_players()
        u.traffic(**self.traffic_args, new_send=False)

    def test_qp_ex_srd_old_send_imm(self):
        self.create_players()
        u.traffic(**self.traffic_args, new_send=False, send_op=e.IBV_WR_SEND_WITH_IMM)

    def test_qp_ex_srd_zero_size(self):
        self.create_players(e.IBV_QP_EX_WITH_SEND)
        self.client.msg_size = 0
        self.server.msg_size = 0
        u.traffic(**self.traffic_args, new_send=True, send_op=e.IBV_WR_SEND)

    def test_post_receive_qp_state_bad_flow(self):
        self.create_players(e.IBV_QP_EX_WITH_SEND, qp_count=1)
        u.post_rq_state_bad_flow(self)

    def test_post_send_qp_state_bad_flow(self):
        self.create_players(e.IBV_QP_EX_WITH_SEND, qp_count=1)
        u.post_sq_state_bad_flow(self)

    def test_full_rq_bad_flow(self):
        self.create_players(e.IBV_QP_EX_WITH_SEND, qp_count=1)
        u.full_rq_bad_flow(self)

    def test_full_sq_bad_flow(self):
        self.create_players(e.IBV_QP_EX_WITH_SEND, qp_count=1)
        self.full_sq_bad_flow()

    def test_rq_with_larger_sgl_bad_flow(self):
        self.create_players(e.IBV_QP_EX_WITH_SEND, qp_count=1)
        u.create_rq_with_larger_sgl_bad_flow(self)



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2020 Nvidia, Inc. All rights reserved. See COPYING file

from libc.stdlib cimport calloc, free

from pyverbs.providers.mlx5.mlx5dv_flow cimport Mlx5FlowMatchParameters
from pyverbs.base import PyverbsRDMAErrno, PyverbsRDMAError
from pyverbs.providers.mlx5.dr_matcher cimport DrMatcher
from pyverbs.providers.mlx5.dr_action cimport DrAction
from pyverbs.pyverbs_error import PyverbsError


cdef class DrRule(PyverbsCM):
    def __init__(self, DrMatcher matcher, Mlx5FlowMatchParameters value,
                 actions=None):
        """
        Initialize DrRule object over underlying mlx5dv_dr_rule C object.
        :param matcher: A matcher with the fields to match on
        :param value: Match parameters with values to match on
        :param actions: List of actions to perform
        """
        super().__init__()
        cdef dv.mlx5dv_dr_action**actions_arr
        actions = [] if actions is None else actions
        actions_arr = <dv.mlx5dv_dr_action**>calloc(len(actions),
                                                    sizeof(dv.mlx5dv_dr_action*))
        if actions_arr == NULL:
            raise PyverbsError('Failed to allocate memory.')
        for i in range(0, len(actions)):
            actions_arr[i] = <dv.mlx5dv_dr_action*>(<DrAction>actions[i]).action
        self.rule = dv.mlx5dv_dr_rule_create(matcher.matcher, value.params,
                                             len(actions), actions_arr)
        free(actions_arr)
        if self.rule == NULL:
            raise PyverbsRDMAErrno('DrRule creation failed.')
        for i in range(0, len(actions)):
            (<DrAction>actions[i]).add_ref(self)
        matcher.add_ref(self)
        self.dr_matcher = matcher

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.rule != NULL:
            if self.logger:
                self.logger.debug('Closing DrRule.')
            rc = dv.mlx5dv_dr_rule_destroy(self.rule)
            if rc:
                raise PyverbsRDMAError('Failed to destroy DrRule.', rc)
            self.rule = NULL
            self.dr_matcher = None



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2020 Nvidia, Inc. All rights reserved. See COPYING file

from pyverbs.pyverbs_error import PyverbsUserError, PyverbsRDMAError
cimport pyverbs.providers.mlx5.libmlx5 as dv
from pyverbs.base import PyverbsRDMAErrno
from pyverbs.pd cimport PD


cdef class Mlx5SigCrc(PyverbsObject):
    def __init__(self, crc_type=0, seed=0):
        """
        Initializes a Mlx5SigCrc object representing mlx5dv_sig_crc C struct.
        :param crc_type: The specific CRC type.
        :param seed: A seed for the CRC calculation per block.
        """
        self.mlx5dv_sig_crc.type = crc_type
        self.mlx5dv_sig_crc.seed = seed


cdef class Mlx5SigT10Dif(PyverbsObject):
    def __init__(self, bg_type=0, bg=0, app_tag=0, ref_tag=0, flags=0):
        """
        Initializes a Mlx5SigT10Dif object representing mlx5dv_sig_t10dif
        C struct.
        :param bg_type: The block guard type to be used.
        :param bg: A seed for the block guard calculation per block.
        :param app_tag: An application tag to generate or validate.
        :param ref_tag: A reference tag to generate or validate.
        :param flags: Flags for the T10DIF attributes.
        """
        self.mlx5dv_sig_t10dif.bg_type = bg_type
        self.mlx5dv_sig_t10dif.bg = bg
        self.mlx5dv_sig_t10dif.app_tag = app_tag
        self.mlx5dv_sig_t10dif.ref_tag = ref_tag
        self.mlx5dv_sig_t10dif.flags = flags


cdef class Mlx5SigBlockDomain(PyverbsObject):
    def __init__(self, sig_type=0, Mlx5SigT10Dif dif=None,
                 Mlx5SigCrc crc=None, block_size=0, comp_mask=0):
        """
        Initializes a Mlx5SigBlockDomain object representing
        mlx5dv_sig_block_domain C struct.
        :param sig_type: The signature type for this domain.
        :param dif: Mlx5SigT10Dif object.
        :param crc: Mlx5SigCrc object.
        :param block_size: The block size for this domain.
        :param comp_mask: Compatibility mask.
        """
        self.mlx5dv_sig_block_domain.sig_type = sig_type
        self.mlx5dv_sig_block_domain.block_size = block_size
        self.mlx5dv_sig_block_domain.comp_mask = comp_mask
        if dif:
            self.mlx5dv_sig_block_domain.sig.dif = &dif.mlx5dv_sig_t10dif
        if crc:
            self.mlx5dv_sig_block_domain.sig.crc = &crc.mlx5dv_sig_crc


cdef class Mlx5SigBlockAttr(PyverbsObject):
    def __init__(self, Mlx5SigBlockDomain mem=None, Mlx5SigBlockDomain wire=None,
                 flags=0, check_mask=0, copy_mask=0, comp_mask=0):
        """
        Initializes a Mlx5SigBlockAttr object representing
        mlx5dv_sig_block_attr C struct.
        :param mem: Mlx5SigBlockDomain of the signature configuration for the
                    memory domain or None if the domain does not have a
                    signature.
        :param wire: Mlx5SigBlockDomain of the signature configuration for the
                     wire domain or None if the domain does not have a
                     signature.
        :param flags: Flags for the block signature attributes.
        :param check_mask: Byte of the input signature is checked if
                           corresponding bit in check_mask is set.
        :param copy_mask: Byte of the signature is copied from the source
                          domain to the destination domain if corresponding
                          bit in copy_mask is set.
        :param comp_mask: Compatibility mask.
        """
        self.mlx5dv_sig_block_attr.flags = flags
        self.mlx5dv_sig_block_attr.check_mask = check_mask
        self.mlx5dv_sig_block_attr.copy_mask = copy_mask
        self.mlx5dv_sig_block_attr.comp_mask = comp_mask
        if mem:
            self.mlx5dv_sig_block_attr.mem = &mem.mlx5dv_sig_block_domain
        if wire:
            self.mlx5dv_sig_block_attr.wire = &wire.mlx5dv_sig_block_domain


cdef class Mlx5SigErr(PyverbsObject):
    def __init__(self, actual_value=0, expected_value=0, offset=0):
        """
        Initializes a Mlx5SigBlockAttr object representing
        mlx5dv_sig_block_attr C struct.
        :param actual_value: The actual value that was calculated from the
                             transferred data.
        :param expected_value: The expected value based on what appears in the
                               signature respected field.
        :param offset: The offset within the transfer where the error happened.
        """
        self.mlx5dv_sig_err.actual_value = actual_value
        self.mlx5dv_sig_err.expected_value = expected_value
        self.mlx5dv_sig_err.offset = offset

    @property
    def actual_value(self):
        return self.mlx5dv_sig_err.actual_value

    @property
    def expected_value(self):
        return self.mlx5dv_sig_err.expected_value

    @property
    def offset(self):
        return self.mlx5dv_sig_err.offset


cdef class Mlx5MkeyErr(PyverbsObject):
    def __init__(self, Mlx5SigErr sig_err=Mlx5SigErr(),
                 err_type=dv.MLX5DV_MKEY_NO_ERR):
        """
        Initializes a Mlx5MkeyErr object representing mlx5dv_mkey_err
        C struct.
        :param sig_err: Mlx5SigErr object that handle the sig error.
        :param err_type: Indicate what kind of error happened.
        """
        self.mlx5dv_mkey_err.err_type = err_type
        self.mlx5dv_mkey_err.err.sig = sig_err.mlx5dv_sig_err

    @property
    def err_type(self):
        return self.mlx5dv_mkey_err.err_type

    @property
    def sig_err(self):
        return Mlx5SigErr(self.mlx5dv_mkey_err.err.sig.actual_value,
                          self.mlx5dv_mkey_err.err.sig.expected_value,
                          self.mlx5dv_mkey_err.err.sig.offset)


cdef class Mlx5MkeyConfAttr(PyverbsObject):
    def __init__(self, conf_flags=0, comp_mask=0):
        """
        Initializes a Mlx5MkeyConfAttr object representing mlx5dv_mkey_conf_attr
        C struct.
        :param conf_flags: Mkey configuration flags.
        :param comp_mask: Compatibility mask.
        """
        self.mlx5dv_mkey_conf_attr.conf_flags = conf_flags
        self.mlx5dv_mkey_conf_attr.comp_mask = comp_mask


cdef class Mlx5MrInterleaved(PyverbsObject):
    def __init__(self, addr, bytes_count, bytes_skip, lkey):
        """
        Initializes a Mlx5MrInterleaved object representing mlx5dv_mr_interleaved
        C struct.
        :param addr: The start of address.
        :param bytes_count: Count of bytes from the address that will hold the
                            real data.
        :param bytes_skip: Count of bytes to skip after the bytes_count.
        :param lkey: The lkey of this memory.
        """
        self.mlx5dv_mr_interleaved.addr = addr
        self.mlx5dv_mr_interleaved.bytes_count = bytes_count
        self.mlx5dv_mr_interleaved.bytes_skip = bytes_skip
        self.mlx5dv_mr_interleaved.lkey = lkey


cdef class Mlx5Mkey(PyverbsCM):
    def __init__(self, PD pd not None, create_flags, max_entries):
        """
        Creates an indirect mkey and store the actual mkey max_entries after the
        mkey creation.
        :param pd: PD instance.
        :param create_flags: Mkey creation flags.
        :param max_entries: Requested max number of pointed entries by this
                            indirect mkey.
        """
        cdef dv.mlx5dv_mkey_init_attr mkey_init
        mkey_init.pd = pd.pd
        mkey_init.create_flags = create_flags
        mkey_init.max_entries = max_entries
        self.mlx5dv_mkey = dv.mlx5dv_create_mkey(&mkey_init)
        if self.mlx5dv_mkey == NULL:
            raise PyverbsRDMAErrno('Failed to create mkey')
        self.max_entries = mkey_init.max_entries
        self.pd = pd
        self.pd.mkeys.add(self)

    def mkey_check(self):
        """
        Checks the mkey for errors and provides the result in err_info on success.
        :return: Mlx5MkeyErr object, the result of the Mkey check.
        """
        mkey_err = Mlx5MkeyErr()
        rc = dv.mlx5dv_mkey_check(self.mlx5dv_mkey, &mkey_err.mlx5dv_mkey_err)
        if rc:
            raise PyverbsRDMAError('Failed to check the mkey', rc)
        return mkey_err

    @property
    def lkey(self):
        return self.mlx5dv_mkey.lkey

    @property
    def rkey(self):
        return self.mlx5dv_mkey.rkey

    @property
    def max_entries(self):
        return self.max_entries

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.mlx5dv_mkey != NULL:
            rc = dv.mlx5dv_destroy_mkey(self.mlx5dv_mkey)
            if rc:
                raise PyverbsRDMAError('Failed to destroy a mkey', rc)
            self.mlx5dv_mkey = NULL
            self.pd = None



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2020 Nvidia, Inc. All rights reserved. See COPYING file

from pyverbs.base import PyverbsRDMAErrno, PyverbsRDMAError
from pyverbs.providers.mlx5.dr_action cimport DrAction
from pyverbs.providers.mlx5.dr_table cimport DrTable
from pyverbs.pyverbs_error import PyverbsError
from pyverbs.base cimport close_weakrefs
from pyverbs.device cimport Context
cimport pyverbs.libibverbs as v
cimport libc.stdio as s
import weakref


cdef class DrDomain(PyverbsCM):
    def __init__(self, Context context, domain_type):
        """
        Initialize DrDomain object over underlying mlx5dv_dr_domain C object.
        :param context: Context object
        :param domain_type: Type of the domain
        """
        super().__init__()
        self.domain = dv.mlx5dv_dr_domain_create(<v.ibv_context*>context.context, domain_type)
        if self.domain == NULL:
            raise PyverbsRDMAErrno('DrDomain creation failed.')
        self.dr_tables = weakref.WeakSet()
        self.context = context
        context.dr_domains.add(self)
        self.dr_actions = weakref.WeakSet()

    def allow_duplicate_rules(self, allow):
        """
        Allows or prevents duplicate rules insertion, by default this feature is
        allowed.
        :param allow: Boolean to allow or prevent
        """
        dv.mlx5dv_dr_domain_allow_duplicate_rules(self.domain, allow)

    cdef add_ref(self, obj):
        if isinstance(obj, DrTable):
            self.dr_tables.add(obj)
        elif isinstance(obj, DrAction):
            self.dr_actions.add(obj)
        else:
            raise PyverbsError('Unrecognized object type')

    def sync(self, flags=dv.MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW |
                         dv.MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW):
        """
        Sync is used in order to flush the rule submission queue.
        :param flags:
            MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW - block until completion of all
                                             software queued tasks
            MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW - clear the steering HW cache to
                                             enforce next packet hits the
                                             latest rules.
            MLX5DV_DR_DOMAIN_SYNC_FLAGS_MEM - sync device memory to free cached
                                              memory.
        """
        if dv.mlx5dv_dr_domain_sync(self.domain, flags):
            raise PyverbsRDMAErrno('DrDomain sync failed.')

    def dump(self, filepath):
        """
        Dumps the debug info of the domain into a file.
        :param filepath: Path to the file
        """
        cdef s.FILE *fp
        fp = s.fopen(filepath.encode('utf-8'), 'w+')
        if fp == NULL:
            raise PyverbsError('Opening dump file failed.')
        rc = dv.mlx5dv_dump_dr_domain(fp, self.domain)
        if rc != 0:
            raise PyverbsRDMAError('Domain dump failed.', rc)
        if s.fclose(fp) != 0:
            raise PyverbsError('Closing dump file failed.')

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.domain != NULL:
            if self.logger:
                self.logger.debug('Closing DrDomain.')
            close_weakrefs([self.dr_actions, self.dr_tables])
            rc = dv.mlx5dv_dr_domain_destroy(self.domain)
            if rc:
                raise PyverbsRDMAError('Failed to destroy DrDomain.', rc)
            self.domain = NULL



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)

#cython: language_level=3


cdef extern from 'infiniband/efadv.h':

    cpdef enum:
        EFADV_DEVICE_ATTR_CAPS_RDMA_READ
        EFADV_DEVICE_ATTR_CAPS_RNR_RETRY
        EFADV_DEVICE_ATTR_CAPS_CQ_WITH_SGID
        EFADV_DEVICE_ATTR_CAPS_RDMA_WRITE

    cpdef enum:
        EFADV_QP_DRIVER_TYPE_SRD

    cpdef enum:
        EFADV_WC_EX_WITH_SGID

    cpdef enum:
        EFADV_MR_ATTR_VALIDITY_RECV_IC_ID
        EFADV_MR_ATTR_VALIDITY_RDMA_READ_IC_ID
        EFADV_MR_ATTR_VALIDITY_RDMA_RECV_IC_ID



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2021 Nvidia, Inc. All rights reserved. See COPYING file

from pyverbs.providers.mlx5.mlx5dv import Wqe, WqeDataSeg, WqeCtrlSeg
from pyverbs.pyverbs_error import PyverbsError
import pyverbs.providers.mlx5.mlx5_enums as dve
from tests.mlx5_base import Mlx5RDMATestCase, Mlx5RcResources
from pyverbs.qp import QPCap
from pyverbs.wr import SGE
import pyverbs.enums as e
import tests.utils as u


class Mlx5RawWqeResources(Mlx5RcResources):
    def create_send_ops_flags(self):
        self.dv_send_ops_flags = dve.MLX5DV_QP_EX_WITH_RAW_WQE
        self.send_ops_flags = e.IBV_QP_EX_WITH_SEND

    def create_qp_cap(self):
        """
        Create QPCap such that work queue elements will wrap around the send
        work queue, this happens due to the iteration count being higher
        than the max_send_wr.
        :return:
        """
        return QPCap(max_send_wr=1, max_recv_wr=4, max_recv_sge=2, max_send_sge=2)


class RawWqeTest(Mlx5RDMATestCase):
    def setUp(self):
        super().setUp()
        self.iters = 10
        self.server = None
        self.client = None

    def prepare_send_elements(self):
        mr = self.client.mr
        sge_count = 2
        unit_size = mr.length / 2
        data_segs = [WqeDataSeg(unit_size, mr.lkey, mr.buf + i * unit_size) for
                     i in range(sge_count)]
        ctrl_seg = WqeCtrlSeg()
        ctrl_seg.fm_ce_se = dve.MLX5_WQE_CTRL_CQ_UPDATE
        segment_num = 1 + len(data_segs)
        ctrl_seg.opmod_idx_opcode = dve.MLX5_OPCODE_SEND
        ctrl_seg.qpn_ds = segment_num | int(self.client.qp.qp_num) << 8
        self.raw_send_wqe = Wqe([ctrl_seg] + data_segs)
        self.regular_send_sge = SGE(mr.buf, mr.length, mr.lkey)

    def mixed_traffic(self):
        s_recv_wr = u.get_recv_wr(self.server)
        u.post_recv(self.server, s_recv_wr)
        self.prepare_send_elements()

        for i in range(self.iters):
            self.client.qp.wr_start()
            if i % 2:
                self.client.mr.write('c' * self.client.mr.length, self.client.mr.length)
                self.client.qp.wr_flags = e.IBV_SEND_SIGNALED
                self.client.qp.wr_send()
                self.client.qp.wr_set_sge(self.regular_send_sge)
            else:
                self.client.mr.write('s' * self.client.mr.length, self.client.mr.length)
                self.client.qp.wr_raw_wqe(self.raw_send_wqe)
            self.client.qp.wr_complete()
            u.poll_cq_ex(self.client.cq)
            u.poll_cq_ex(self.server.cq)
            u.post_recv(self.server, s_recv_wr)
            expected_opcode = e.IBV_WC_SEND if i % 2 else e.IBV_WC_DRIVER2

            if self.client.cq.read_opcode() != expected_opcode:
                raise PyverbsError('Opcode validation failed: expected '
                                   f'{expected_opcode}, received {self.client.cq.read_opcode()}')

            act_buffer = self.server.mr.read(self.server.mr.length, 0)
            u.validate(act_buffer, i % 2, self.server.mr.length)

    def test_mixed_raw_wqe_traffic(self):
        """
        Runs traffic with a mix of SEND opcode regular WQEs and SEND opcode RAW
        WQEs.
        """
        self.create_players(Mlx5RawWqeResources)
        self.mixed_traffic()



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2020 Nvidia, Inc. All rights reserved. See COPYING file

from pyverbs.providers.mlx5.mlx5dv_flow cimport Mlx5FlowMatchParameters
from pyverbs.pyverbs_error import PyverbsError, PyverbsRDMAError
from pyverbs.providers.mlx5.dr_table cimport DrTable
from pyverbs.providers.mlx5.dr_rule cimport DrRule
from pyverbs.base import PyverbsRDMAErrno
from pyverbs.base cimport close_weakrefs
import weakref


cdef class DrMatcher(PyverbsCM):
    def __init__(self, DrTable table, priority, match_criteria_enable,
                  Mlx5FlowMatchParameters mask):
        """
        Initialize DrMatcher object over underlying mlx5dv_dr_matcher C object.
        :param table: Table to create the matcher on
        :param priority: Matcher priority
        :param match_criteria_enable: Bitmask representing which of the
                                      headers and parameters in match_criteria
                                      are used in defining the Flow.
                                      Bit 0: outer_headers
                                      Bit 1: misc_parameters
                                      Bit 2: inner_headers
                                      Bit 3: misc_parameters_2
                                      Bit 4: misc_parameters_3
                                      Bit 5: misc_parameters_4
        :param mask: Match parameters to match on
        """
        super().__init__()
        self.matcher = dv.mlx5dv_dr_matcher_create(table.table, priority,
                                                   match_criteria_enable,
                                                   mask.params)
        if self.matcher == NULL:
            raise PyverbsRDMAErrno('DrMatcher creation failed.')
        table.add_ref(self)
        self.dr_table = table
        self.dr_rules = weakref.WeakSet()

    cdef add_ref(self, obj):
        if isinstance(obj, DrRule):
            self.dr_rules.add(obj)
        else:
            raise PyverbsError('Unrecognized object type')

    def set_layout(self, log_num_of_rules=0, flags=dv.MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE):
        """
        Set the size of the table for the matcher
        :param log_num_of_rules: Log of the table size (relevant for MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE)
        :param flags: Matcher layout flags
        """
        cdef dv.mlx5dv_dr_matcher_layout matcher_layout
        matcher_layout.log_num_of_rules_hint = log_num_of_rules
        matcher_layout.flags = flags
        rc = dv.mlx5dv_dr_matcher_set_layout(self.matcher, &matcher_layout)
        if rc:
            raise PyverbsRDMAError('Setting matcher layout failed.', rc)

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.matcher != NULL:
            if self.logger:
                self.logger.debug('Closing Matcher.')
            close_weakrefs([self.dr_rules])
            if dv.mlx5dv_dr_matcher_destroy(self.matcher):
                raise PyverbsRDMAErrno('Failed to destroy DrMatcher.')
            self.matcher = NULL
            self.dr_table = None



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2020 Nvidia, Inc. All rights reserved. See COPYING file

from pyverbs.base import PyverbsRDMAErrno, PyverbsRDMAError
from pyverbs.providers.mlx5.dr_matcher import DrMatcher
from pyverbs.providers.mlx5.dr_domain cimport DrDomain
from pyverbs.providers.mlx5.dr_action cimport DrAction
from pyverbs.pyverbs_error import PyverbsError
from pyverbs.base cimport close_weakrefs
import weakref


cdef class DrTable(PyverbsCM):
    def __init__(self, DrDomain domain, level):
        """
        Initialize DrTable object over underlying mlx5dv_dr_table C object.
        :param domain: Domain object
        :param level: Table level
        """
        super().__init__()
        self.table = dv.mlx5dv_dr_table_create(domain.domain, level)
        if self.table == NULL:
            raise PyverbsRDMAErrno('DrTable creation failed.')
        domain.add_ref(self)
        self.dr_domain = domain
        self.dr_matchers = weakref.WeakSet()
        self.dr_actions = weakref.WeakSet()

    cdef add_ref(self, obj):
        if isinstance(obj, DrMatcher):
            self.dr_matchers.add(obj)
        elif isinstance(obj, DrAction):
            self.dr_actions.add(obj)
        else:
            raise PyverbsError('Unrecognized object type')

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.table != NULL:
            if self.logger:
                self.logger.debug('Closing DrTable.')
            close_weakrefs([self.dr_matchers, self.dr_actions])
            rc = dv.mlx5dv_dr_table_destroy(self.table)
            if rc:
                raise PyverbsRDMAError('Failed to destroy DrTable.', rc)
            self.table = NULL
            self.dr_domain = None
            self.dr_actions = None



#!/usr/bin/env python3
# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2018, Mellanox Technologies. All rights reserved.  See COPYING file

from args_parser import parser
import unittest
import os
from importlib.machinery import SourceFileLoader


module_path = os.path.join(os.path.dirname(__file__), '__init__.py')
tests = SourceFileLoader('tests', module_path).load_module()
parser.parse_args()
unittest.main(module=tests)



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019 Mellanox Technologies, Inc . All rights reserved. See COPYING file

import abc

from pyverbs.cmid import CMID, AddrInfo, CMEventChannel, ConnParam, UDParam
from pyverbs.qp import QPCap, QPInitAttr, QPAttr, QP
from pyverbs.pyverbs_error import PyverbsUserError
import pyverbs.cm_enums as ce
import pyverbs.enums as e
from pyverbs.cq import CQ


GRH_SIZE = 40
qp_type_per_ps = {ce.RDMA_PS_TCP: e.IBV_QPT_RC, ce.RDMA_PS_UDP: e.IBV_QPT_UD,
                  ce.RDMA_PS_IPOIB : e.IBV_QPT_UD}


class CMResources(abc.ABC):
    """
    CMResources class is an abstract base class which contains basic resources
    for RDMA CM communication.
    """
    def __init__(self, addr=None, passive=None, **kwargs):
        """
        :param addr: Local address to bind to.
        :param passive: Indicate if this CM is the passive CM.
        :param kwargs: Arguments:
            * *port* (str)
                Port number of the address
            * *with_ext_qp* (bool)
                If set, an external RC QP will be created and used by RDMACM
            * *port_space* (str)
                If set, indicates the CMIDs port space
        """
        self.qp_init_attr = None
        self.passive = passive
        self.with_ext_qp = kwargs.get('with_ext_qp', False)
        self.port = kwargs.get('port') if kwargs.get('port') else '7471'
        self.ib_port = int(kwargs.get('ib_port', '1'))
        self.port_space = kwargs.get('port_space', ce.RDMA_PS_TCP)
        self.remote_operation = kwargs.get('remote_op')
        self.qp_type = qp_type_per_ps[self.port_space]
        self.qp_init_attr = QPInitAttr(qp_type=self.qp_type, cap=QPCap())
        self.connected = False
        # When passive side (server) listens to incoming connection requests,
        # for each new request it creates a new cmid which is used to establish
        # the connection with the remote side
        self.msg_size = 1024
        self.num_msgs = 10
        self.channel = None
        self.cq = None
        self.qps = {}
        self.mr = None
        self.remote_qpn = None
        self.ud_params = None
        self.child_ids = {}
        self.cmids = {}
        if self.passive:
            self.ai = AddrInfo(src=addr, src_service=self.port,
                               port_space=self.port_space, flags=ce.RAI_PASSIVE)
        else:
            self.ai = AddrInfo(src=addr, dst=addr, dst_service=self.port,
                               port_space=self.port_space)

    @property
    def child_id(self):
        if self.child_ids:
            return self.child_ids[0]

    @property
    def cmid(self):
        if self.cmids:
            return self.cmids[0]

    @property
    def qp(self):
        if self.qps:
            return self.qps[0]

    def create_mr(self):
        cmid = self.child_id if self.passive else self.cmid
        mr_remote_function = {None: cmid.reg_msgs, 'read': cmid.reg_read,
                              'write': cmid.reg_write}
        self.mr = mr_remote_function[self.remote_operation](self.msg_size + GRH_SIZE)

    def create_event_channel(self):
        self.channel = CMEventChannel()

    def create_qp_init_attr(self, rcq=None, scq=None):
        return QPInitAttr(qp_type=self.qp_type, rcq=rcq, scq=scq,
                          cap=QPCap(max_recv_wr=1))

    def create_conn_param(self, qp_num=0, conn_idx=0):
        if self.with_ext_qp:
            qp_num = self.qp.qp_num
        return ConnParam(qp_num=qp_num)

    def set_ud_params(self, cm_event):
        if self.port_space in [ce.RDMA_PS_UDP, ce.RDMA_PS_IPOIB]:
            self.ud_params = UDParam(cm_event)

    def my_qp_number(self):
        if self.with_ext_qp:
            return self.qp.qp_num
        else:
            cm = self.child_id if self.passive else self.cmid
            return cm.qpn

    def create_qp(self, conn_idx=0):
        """
        Create an rdmacm QP. If self.with_ext_qp is set, then an external CQ and
        QP will be created. In case that CQ is already created, it is used
        for the newly created QP.
        :param conn_idx: The connection index.
        """
        cmid = self.child_id if self.passive else self.cmid
        if not self.with_ext_qp:
            cmid.create_qp(self.create_qp_init_attr())
        else:
            self.create_cq(cmid)
            init_attr = self.create_qp_init_attr(rcq=self.cq, scq=self.cq)
            self.qps[conn_idx] = QP(cmid.pd, init_attr, QPAttr())

    def create_cq(self, cmid):
        if not self.cq:
            self.cq = CQ(cmid.context, self.num_msgs, None, None, 0)

    def modify_ext_qp_to_rts(self, conn_idx=0):
        cmid = self.child_id if self.passive else self.cmid
        attr, mask = cmid.init_qp_attr(e.IBV_QPS_INIT)
        self.qps[conn_idx].modify(attr, mask)
        attr, mask = cmid.init_qp_attr(e.IBV_QPS_RTR)
        self.qps[conn_idx].modify(attr, mask)
        attr, mask = cmid.init_qp_attr(e.IBV_QPS_RTS)
        self.qps[conn_idx].modify(attr, mask)

    def mem_write(self, data, size, offset=0):
        self.mr.write(data, size, offset)

    def mem_read(self, size=None, offset=0):
        size_ = self.msg_size if size is None else size
        return self.mr.read(size_, offset)

    @abc.abstractmethod
    def create_child_id(self, cm_event=None):
        pass

    @property
    def mr_lkey(self):
        return self.mr.lkey


class AsyncCMResources(CMResources):
    """
    AsyncCMResources class contains resources for RDMA CM asynchronous
    communication.
    :param addr: Local address to bind to.
    :param passive: Indicate if this CM is the passive CM.
    """
    def __init__(self, addr=None, passive=None, **kwargs):
        super(AsyncCMResources, self).__init__(addr=addr, passive=passive,
                                               **kwargs)
        self.create_event_channel()

    def create_cmid(self, idx=0):
        self.cmids[idx] = CMID(creator=self.channel, port_space=self.port_space)

    def create_child_id(self, cm_event=None):
        if not self.passive:
            raise PyverbsUserError('create_child_id can be used only in passive side')
        new_child_idx = len(self.child_ids)
        self.child_ids[new_child_idx] = CMID(creator=cm_event, listen_id=self.cmid)


class SyncCMResources(CMResources):
    """
    SyncCMResources class contains resources for RDMA CM synchronous
    communication.
    :param addr: Local address to bind to.
    :param passive: Indicate if this CM is the passive CM.
    """
    def __init__(self, addr=None, passive=None, **kwargs):
        super(SyncCMResources, self).__init__(addr=addr, passive=passive,
                                              **kwargs)

    def create_cmid(self, idx=0):
        self.cmids[idx] = CMID(creator=self.ai, qp_init_attr=self.qp_init_attr)

    def create_child_id(self, cm_event=None):
        if not self.passive:
            raise PyverbsUserError('create_child_id can be used only in passive side')
        new_child_idx = len(self.child_ids)
        self.child_ids[new_child_idx] = self.cmid.get_request()



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2021 Nvidia, Inc. All rights reserved. See COPYING file

#cython: language_level=3

from pyverbs.base cimport PyverbsObject, PyverbsCM
cimport pyverbs.providers.mlx5.libmlx5 as dv
from pyverbs.device cimport Context
from pyverbs.pd cimport PD


cdef class Mlx5CryptoLoginAttr(PyverbsObject):
    cdef dv.mlx5dv_crypto_login_attr mlx5dv_crypto_login_attr

cdef class Mlx5CryptoExtLoginAttr(PyverbsObject):
    cdef dv.mlx5dv_crypto_login_attr_ex mlx5dv_crypto_login_attr_ex
    cdef object credential

cdef class Mlx5DEKInitAttr(PyverbsObject):
    cdef dv.mlx5dv_dek_init_attr mlx5dv_dek_init_attr
    cdef PD pd

cdef class Mlx5DEKAttr(PyverbsObject):
    cdef dv.mlx5dv_dek_attr mlx5dv_dek_attr

cdef class Mlx5CryptoAttr(PyverbsObject):
    cdef dv.mlx5dv_crypto_attr mlx5dv_crypto_attr

cdef class Mlx5DEK(PyverbsCM):
    cdef dv.mlx5dv_dek *mlx5dv_dek
    cdef PD pd

cdef class Mlx5CryptoLogin(PyverbsCM):
    cdef dv.mlx5dv_crypto_login_obj *crypto_login_obj
    cdef Context context



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2019, Mellanox Technologies. All rights reserved.

#cython: language_level=3

from pyverbs.base cimport PyverbsObject
from pyverbs.device cimport Context
cimport pyverbs.libibverbs as v
from .base cimport PyverbsCM


cdef class PD(PyverbsCM):
    cdef v.ibv_pd *pd
    cdef Context ctx
    cdef add_ref(self, obj)
    cdef remove_ref(self, obj)
    cdef object srqs
    cdef object mrs
    cdef object mws
    cdef object ahs
    cdef object qps
    cdef object parent_domains
    cdef object wqs
    cdef object mkeys
    cdef object deks
    cdef object _is_imported

cdef class ParentDomainInitAttr(PyverbsObject):
    cdef v.ibv_parent_domain_init_attr init_attr
    cdef object pd
    cdef object alloc
    cdef object dealloc

cdef class ParentDomain(PD):
    cdef add_ref(self, obj)
    cdef object protection_domain
    cdef object cqs

cdef class ParentDomainContext(PyverbsObject):
    cdef object p_alloc
    cdef object p_free
    cdef object pd
    cdef object user_data
